import numpy as np
import torch
import torch.nn as nn
from mmcv.cnn import Conv2d, Linear, MaxPool2d, kaiming_init, normal_init
from mmcv.runner import force_fp32
from torch.nn.modules.utils import _pair

from mmdet.models.builder import HEADS, build_loss


@HEADS.register_module()
class MaskIoUHead(nn.Module):
    """Mask IoU Head.

    This head predicts the IoU of predicted masks and corresponding gt masks.
    """

    def __init__(self,
                 num_convs=4,
                 num_fcs=2,
                 roi_feat_size=14,
                 in_channels=256,
                 conv_out_channels=256,
                 fc_out_channels=1024,
                 num_classes=80,
                 loss_iou=dict(type='MSELoss', loss_weight=0.5)):
        super(MaskIoUHead, self).__init__()
        self.in_channels = in_channels
        self.conv_out_channels = conv_out_channels
        self.fc_out_channels = fc_out_channels
        self.num_classes = num_classes
        self.fp16_enabled = False

        self.convs = nn.ModuleList()
        for i in range(num_convs):
            if i == 0:
                # concatenation of mask feature and mask prediction
                in_channels = self.in_channels + 1
            else:
                in_channels = self.conv_out_channels
            stride = 2 if i == num_convs - 1 else 1
            self.convs.append(
                Conv2d(
                    in_channels,
                    self.conv_out_channels,
                    3,
                    stride=stride,
                    padding=1))

        roi_feat_size = _pair(roi_feat_size)
        pooled_area = (roi_feat_size[0] // 2) * (roi_feat_size[1] // 2)
        self.fcs = nn.ModuleList()
        for i in range(num_fcs):
            in_channels = (
                self.conv_out_channels *
                pooled_area if i == 0 else self.fc_out_channels)
            self.fcs.append(Linear(in_channels, self.fc_out_channels))

        self.fc_mask_iou = Linear(self.fc_out_channels, self.num_classes)
        self.relu = nn.ReLU()
        self.max_pool = MaxPool2d(2, 2)
        self.loss_iou = build_loss(loss_iou)

    def init_weights(self):
        for conv in self.convs:
            kaiming_init(conv)
        for fc in self.fcs:
            kaiming_init(
                fc,
                a=1,
                mode='fan_in',
                nonlinearity='leaky_relu',
                distribution='uniform')
        normal_init(self.fc_mask_iou, std=0.01)

    def forward(self, mask_feat, mask_pred):
        mask_pred = mask_pred.sigmoid()
        mask_pred_pooled = self.max_pool(mask_pred.unsqueeze(1))

        x = torch.cat((mask_feat, mask_pred_pooled), 1)

        for conv in self.convs:
            x = self.relu(conv(x))
        x = x.flatten(1)
        for fc in self.fcs:
            x = self.relu(fc(x))
        mask_iou = self.fc_mask_iou(x)
        return mask_iou

    @force_fp32(apply_to=('mask_iou_pred', ))
    def loss(self, mask_iou_pred, mask_iou_targets):
        pos_inds = mask_iou_targets > 0
        if pos_inds.sum() > 0:
            loss_mask_iou = self.loss_iou(mask_iou_pred[pos_inds],
                                          mask_iou_targets[pos_inds])
        else:
            loss_mask_iou = mask_iou_pred.sum() * 0
        return dict(loss_mask_iou=loss_mask_iou)

    @force_fp32(apply_to=('mask_pred', ))
    def get_targets(self, sampling_results, gt_masks, mask_pred, mask_targets,
                    rcnn_train_cfg):
        """Compute target of mask IoU.

        Mask IoU target is the IoU of the predicted mask (inside a bbox) and
        the gt mask of corresponding gt mask (the whole instance).
        The intersection area is computed inside the bbox, and the gt mask area
        is computed with two steps, firstly we compute the gt area inside the
        bbox, then divide it by the area ratio of gt area inside the bbox and
        the gt area of the whole instance.

        Args:
            sampling_results (list[:obj:`SamplingResult`]): sampling results.
            gt_masks (BitmapMask | PolygonMask): Gt masks (the whole instance)
                of each image, with the same shape of the input image.
            mask_pred (Tensor): Predicted masks of each positive proposal,
                shape (num_pos, h, w).
            mask_targets (Tensor): Gt mask of each positive proposal,
                binary map of the shape (num_pos, h, w).
            rcnn_train_cfg (dict): Training config for R-CNN part.

        Returns:
            Tensor: mask iou target (length == num positive).
        """
        pos_proposals = [res.pos_bboxes for res in sampling_results]
        pos_assigned_gt_inds = [
            res.pos_assigned_gt_inds for res in sampling_results
        ]

        # compute the area ratio of gt areas inside the proposals and
        # the whole instance
        area_ratios = map(self._get_area_ratio, pos_proposals,
                          pos_assigned_gt_inds, gt_masks)
        area_ratios = torch.cat(list(area_ratios))
        assert mask_targets.size(0) == area_ratios.size(0)

        mask_pred = (mask_pred > rcnn_train_cfg.mask_thr_binary).float()
        mask_pred_areas = mask_pred.sum((-1, -2))

        # mask_pred and mask_targets are binary maps
        overlap_areas = (mask_pred * mask_targets).sum((-1, -2))

        # compute the mask area of the whole instance
        gt_full_areas = mask_targets.sum((-1, -2)) / (area_ratios + 1e-7)

        mask_iou_targets = overlap_areas / (
            mask_pred_areas + gt_full_areas - overlap_areas)
        return mask_iou_targets

    def _get_area_ratio(self, pos_proposals, pos_assigned_gt_inds, gt_masks):
        """Compute area ratio of the gt mask inside the proposal and the gt
        mask of the corresponding instance."""
        num_pos = pos_proposals.size(0)
        if num_pos > 0:
            area_ratios = []
            proposals_np = pos_proposals.cpu().numpy()
            pos_assigned_gt_inds = pos_assigned_gt_inds.cpu().numpy()
            # compute mask areas of gt instances (batch processing for speedup)
            gt_instance_mask_area = gt_masks.areas
            for i in range(num_pos):
                gt_mask = gt_masks[pos_assigned_gt_inds[i]]

                # crop the gt mask inside the proposal
                bbox = proposals_np[i, :].astype(np.int32)
                gt_mask_in_proposal = gt_mask.crop(bbox)

                ratio = gt_mask_in_proposal.areas[0] / (
                    gt_instance_mask_area[pos_assigned_gt_inds[i]] + 1e-7)
                area_ratios.append(ratio)
            area_ratios = torch.from_numpy(np.stack(area_ratios)).float().to(
                pos_proposals.device)
        else:
            area_ratios = pos_proposals.new_zeros((0, ))
        return area_ratios

    @force_fp32(apply_to=('mask_iou_pred', ))
    def get_mask_scores(self, mask_iou_pred, det_bboxes, det_labels):
        """Get the mask scores.

        mask_score = bbox_score * mask_iou
        """
        inds = range(det_labels.size(0))
        mask_scores = mask_iou_pred[inds, det_labels] * det_bboxes[inds, -1]
        mask_scores = mask_scores.cpu().numpy()
        det_labels = det_labels.cpu().numpy()
        return [mask_scores[det_labels == i] for i in range(self.num_classes)]



# Copyright (c) OpenMMLab. All rights reserved.
import torch
from torch import nn
from torch.autograd import Function
from torch.nn.modules.utils import _pair

from ..utils import ext_loader

ext_module = ext_loader.load_ext(
    '_ext', ['dynamic_voxelize_forward', 'hard_voxelize_forward'])


class _Voxelization(Function):

    @staticmethod
    def forward(ctx,
                points,
                voxel_size,
                coors_range,
                max_points=35,
                max_voxels=20000):
        """Convert kitti points(N, >=3) to voxels.

        Args:
            points (torch.Tensor): [N, ndim]. Points[:, :3] contain xyz points
                and points[:, 3:] contain other information like reflectivity.
            voxel_size (tuple or float): The size of voxel with the shape of
                [3].
            coors_range (tuple or float): The coordinate range of voxel with
                the shape of [6].
            max_points (int, optional): maximum points contained in a voxel. if
                max_points=-1, it means using dynamic_voxelize. Default: 35.
            max_voxels (int, optional): maximum voxels this function create.
                for second, 20000 is a good choice. Users should shuffle points
                before call this function because max_voxels may drop points.
                Default: 20000.

        Returns:
            voxels_out (torch.Tensor): Output voxels with the shape of [M,
                max_points, ndim]. Only contain points and returned when
                max_points != -1.
            coors_out (torch.Tensor): Output coordinates with the shape of
                [M, 3].
            num_points_per_voxel_out (torch.Tensor): Num points per voxel with
                the shape of [M]. Only returned when max_points != -1.
        """
        if max_points == -1 or max_voxels == -1:
            coors = points.new_zeros(size=(points.size(0), 3), dtype=torch.int)
            ext_module.dynamic_voxelize_forward(points, coors, voxel_size,
                                                coors_range, 3)
            return coors
        else:
            voxels = points.new_zeros(
                size=(max_voxels, max_points, points.size(1)))
            coors = points.new_zeros(size=(max_voxels, 3), dtype=torch.int)
            num_points_per_voxel = points.new_zeros(
                size=(max_voxels, ), dtype=torch.int)
            voxel_num = ext_module.hard_voxelize_forward(
                points, voxels, coors, num_points_per_voxel, voxel_size,
                coors_range, max_points, max_voxels, 3)
            # select the valid voxels
            voxels_out = voxels[:voxel_num]
            coors_out = coors[:voxel_num]
            num_points_per_voxel_out = num_points_per_voxel[:voxel_num]
            return voxels_out, coors_out, num_points_per_voxel_out


voxelization = _Voxelization.apply


class Voxelization(nn.Module):
    """Convert kitti points(N, >=3) to voxels.

    Please refer to `PVCNN <https://arxiv.org/abs/1907.03739>`_ for more
    details.

    Args:
        voxel_size (tuple or float): The size of voxel with the shape of [3].
        point_cloud_range (tuple or float): The coordinate range of voxel with
            the shape of [6].
        max_num_points (int): maximum points contained in a voxel. if
            max_points=-1, it means using dynamic_voxelize.
        max_voxels (int, optional): maximum voxels this function create.
            for second, 20000 is a good choice. Users should shuffle points
            before call this function because max_voxels may drop points.
            Default: 20000.
    """

    def __init__(self,
                 voxel_size,
                 point_cloud_range,
                 max_num_points,
                 max_voxels=20000):
        super().__init__()

        self.voxel_size = voxel_size
        self.point_cloud_range = point_cloud_range
        self.max_num_points = max_num_points
        if isinstance(max_voxels, tuple):
            self.max_voxels = max_voxels
        else:
            self.max_voxels = _pair(max_voxels)

        point_cloud_range = torch.tensor(
            point_cloud_range, dtype=torch.float32)
        voxel_size = torch.tensor(voxel_size, dtype=torch.float32)
        grid_size = (point_cloud_range[3:] -
                     point_cloud_range[:3]) / voxel_size
        grid_size = torch.round(grid_size).long()
        input_feat_shape = grid_size[:2]
        self.grid_size = grid_size
        # the origin shape is as [x-len, y-len, z-len]
        # [w, h, d] -> [d, h, w]
        self.pcd_shape = [*input_feat_shape, 1][::-1]

    def forward(self, input):
        if self.training:
            max_voxels = self.max_voxels[0]
        else:
            max_voxels = self.max_voxels[1]

        return voxelization(input, self.voxel_size, self.point_cloud_range,
                            self.max_num_points, max_voxels)

    def __repr__(self):
        s = self.__class__.__name__ + '('
        s += 'voxel_size=' + str(self.voxel_size)
        s += ', point_cloud_range=' + str(self.point_cloud_range)
        s += ', max_num_points=' + str(self.max_num_points)
        s += ', max_voxels=' + str(self.max_voxels)
        s += ')'
        return s



from typing import Any, Callable, Dict, List, Optional, Union
from diffusers.models.attention_processor import Attention
from diffusers.models import AutoencoderKL, UNet2DConditionModel
import torch
import numpy as np
import PIL


class Pix2PixZeroAttnProcessor:
    """An attention processor class to store the attention weights.
    In Pix2Pix Zero, it happens during computations in the cross-attention blocks."""

    def __init__(self, is_pix2pix_zero=False):
        self.is_pix2pix_zero = is_pix2pix_zero
        if self.is_pix2pix_zero:
            self.reference_cross_attn_map = {}

    def __call__(
            self,
            attn: Attention,
            hidden_states,
            encoder_hidden_states=None,
            attention_mask=None,
            timestep=None,
            loss=None,
    ):
        batch_size, sequence_length, _ = hidden_states.shape
        attention_mask = attn.prepare_attention_mask(attention_mask, sequence_length, batch_size)
        query = attn.to_q(hidden_states)

        if encoder_hidden_states is None:
            encoder_hidden_states = hidden_states
        elif attn.norm_cross:
            encoder_hidden_states = attn.norm_encoder_hidden_states(encoder_hidden_states)

        key = attn.to_k(encoder_hidden_states)
        value = attn.to_v(encoder_hidden_states)

        query = attn.head_to_batch_dim(query)
        key = attn.head_to_batch_dim(key)
        value = attn.head_to_batch_dim(value)

        attention_probs = attn.get_attention_scores(query, key, attention_mask)
        if self.is_pix2pix_zero and timestep is not None:
            # new bookkeeping to save the attention weights.
            if loss is None:
                self.reference_cross_attn_map[timestep.item()] = attention_probs.detach().cpu()
            # compute loss
            elif loss is not None:
                prev_attn_probs = self.reference_cross_attn_map.pop(timestep.item())
                loss.compute_loss(attention_probs, prev_attn_probs.to(attention_probs.device))

        hidden_states = torch.bmm(attention_probs, value)
        hidden_states = attn.batch_to_head_dim(hidden_states)

        # linear proj
        hidden_states = attn.to_out[0](hidden_states)
        # dropout
        hidden_states = attn.to_out[1](hidden_states)

        return hidden_states


def prepare_unet(unet: UNet2DConditionModel):
    """Modifies the UNet (`unet`) to perform Pix2Pix Zero optimizations."""
    pix2pix_zero_attn_procs = {}
    for name in unet.attn_processors.keys():
        module_name = name.replace(".processor", "")
        module = unet.get_submodule(module_name)
        if "attn2" in name:
            pix2pix_zero_attn_procs[name] = Pix2PixZeroAttnProcessor(is_pix2pix_zero=True)
            module.requires_grad_(True)
        else:
            pix2pix_zero_attn_procs[name] = Pix2PixZeroAttnProcessor(is_pix2pix_zero=False)
            module.requires_grad_(False)

    unet.set_attn_processor(pix2pix_zero_attn_procs)
    return unet


def _in_step(config, step):
    in_step = False
    try:
        start_step = config.start_step
        end_step = config.end_step
        if start_step <= step < end_step:
            in_step = True
    except:
        in_step = False
    return in_step


def _classify_blocks(block_list, name):
    return Any([(block in name) for block in block_list])



# Copyright (c) OpenMMLab. All rights reserved.
import torch.nn as nn

from .registry import PADDING_LAYERS

PADDING_LAYERS.register_module('zero', module=nn.ZeroPad2d)
PADDING_LAYERS.register_module('reflect', module=nn.ReflectionPad2d)
PADDING_LAYERS.register_module('replicate', module=nn.ReplicationPad2d)


def build_padding_layer(cfg, *args, **kwargs):
    """Build padding layer.

    Args:
        cfg (None or dict): The padding layer config, which should contain:
            - type (str): Layer type.
            - layer args: Args needed to instantiate a padding layer.

    Returns:
        nn.Module: Created padding layer.
    """
    if not isinstance(cfg, dict):
        raise TypeError('cfg must be a dict')
    if 'type' not in cfg:
        raise KeyError('the cfg dict must contain the key "type"')

    cfg_ = cfg.copy()
    padding_type = cfg_.pop('type')
    if padding_type not in PADDING_LAYERS:
        raise KeyError(f'Unrecognized padding type {padding_type}.')
    else:
        padding_layer = PADDING_LAYERS.get(padding_type)

    layer = padding_layer(*args, **kwargs, **cfg_)

    return layer



import torch.nn as nn
import torch.nn.functional as F
from mmcv.cnn import ConvModule, kaiming_init
from mmcv.runner import auto_fp16, force_fp32

from mmdet.models.builder import HEADS


@HEADS.register_module()
class FusedSemanticHead(nn.Module):
    r"""Multi-level fused semantic segmentation head.

    .. code-block:: none

        in_1 -> 1x1 conv ---
                            |
        in_2 -> 1x1 conv -- |
                           ||
        in_3 -> 1x1 conv - ||
                          |||                  /-> 1x1 conv (mask prediction)
        in_4 -> 1x1 conv -----> 3x3 convs (*4)
                            |                  \-> 1x1 conv (feature)
        in_5 -> 1x1 conv ---
    """  # noqa: W605

    def __init__(self,
                 num_ins,
                 fusion_level,
                 num_convs=4,
                 in_channels=256,
                 conv_out_channels=256,
                 num_classes=183,
                 ignore_label=255,
                 loss_weight=0.2,
                 conv_cfg=None,
                 norm_cfg=None):
        super(FusedSemanticHead, self).__init__()
        self.num_ins = num_ins
        self.fusion_level = fusion_level
        self.num_convs = num_convs
        self.in_channels = in_channels
        self.conv_out_channels = conv_out_channels
        self.num_classes = num_classes
        self.ignore_label = ignore_label
        self.loss_weight = loss_weight
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.fp16_enabled = False

        self.lateral_convs = nn.ModuleList()
        for i in range(self.num_ins):
            self.lateral_convs.append(
                ConvModule(
                    self.in_channels,
                    self.in_channels,
                    1,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,
                    inplace=False))

        self.convs = nn.ModuleList()
        for i in range(self.num_convs):
            in_channels = self.in_channels if i == 0 else conv_out_channels
            self.convs.append(
                ConvModule(
                    in_channels,
                    conv_out_channels,
                    3,
                    padding=1,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg))
        self.conv_embedding = ConvModule(
            conv_out_channels,
            conv_out_channels,
            1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg)
        self.conv_logits = nn.Conv2d(conv_out_channels, self.num_classes, 1)

        self.criterion = nn.CrossEntropyLoss(ignore_index=ignore_label)

    def init_weights(self):
        kaiming_init(self.conv_logits)

    @auto_fp16()
    def forward(self, feats):
        x = self.lateral_convs[self.fusion_level](feats[self.fusion_level])
        fused_size = tuple(x.shape[-2:])
        for i, feat in enumerate(feats):
            if i != self.fusion_level:
                feat = F.interpolate(
                    feat, size=fused_size, mode='bilinear', align_corners=True)
                x += self.lateral_convs[i](feat)

        for i in range(self.num_convs):
            x = self.convs[i](x)

        mask_pred = self.conv_logits(x)
        x = self.conv_embedding(x)
        return mask_pred, x

    @force_fp32(apply_to=('mask_pred', ))
    def loss(self, mask_pred, labels):
        labels = labels.squeeze(1).long()
        loss_semantic_seg = self.criterion(mask_pred, labels)
        loss_semantic_seg *= self.loss_weight
        return loss_semantic_seg



"""Self-implemented GLIDE pipeline for with source code and basic libraries like PIL, numpy, torch, etc."""

from typing import List, Optional, Tuple, Union

from IPython.display import display
import PIL
from PIL import Image
import numpy as np
import torch
import torch.nn.functional as F

# from diffusers
from diffusers.utils.pil_utils import PIL_INTERPOLATION
from diffusers.pipelines.pipeline_utils import ImagePipelineOutput

from imagen_hub.utils.image_helper import rgba_to_01_mask

from .glide_text2im.download import load_checkpoint  # TODO, make it download online.
from .glide_text2im.model_creation import (
    create_model_and_diffusion,
    model_and_diffusion_defaults,
    model_and_diffusion_defaults_upsampler
)

# from .utils import randn_tensor
# from ..pipeline_utils import DiffusionPipeline, ImagePipelineOutput


def _preprocess_image(image: Union[List, PIL.Image.Image, torch.Tensor], image_size):
    if isinstance(image, torch.Tensor):
        return image
    elif isinstance(image, PIL.Image.Image):
        image = [image]
        # to torch

    if isinstance(image[0], PIL.Image.Image):
        w, h = image_size, image_size  # resize to 64x64 or 256x256
        image = [np.array(i.resize((w, h), resample=Image.BICUBIC))[None, :] for i in image]
        image = np.concatenate(image, axis=0)
        image = np.array(image).astype(np.float32) / 255.0
        image = image.transpose(0, 3, 1, 2)  # [batch, channel, height, width]
        image = 2.0 * image - 1.0
        image = torch.from_numpy(image)
    elif isinstance(image[0], torch.Tensor):
        image = torch.cat(image, dim=0)

    return image


def _preprocess_mask(mask: Union[List, PIL.Image.Image, torch.Tensor], image_size):
    if isinstance(mask, torch.Tensor):
        # print("mask is already a tensor")
        pass
    elif isinstance(mask, PIL.Image.Image):
        # print("mask is a PIL image")
        mask = [mask]

    if isinstance(mask[0], PIL.Image.Image):
        # print("mask[0] shape", mask[0].size)
        # print("mask[1] shape", mask[1].size)
        w, h = image_size, image_size  # resize to 64x64 or 256x256
        # resize
        mask = [m.resize((w, h), resample=PIL_INTERPOLATION["nearest"]) for m in mask]
        # for each mask, do the rgba_to_01_mask
        mask = [rgba_to_01_mask(m, reverse=True) for m in mask]
        # print("mask[0] shape", mask[0].shape)
        # print("mask[1] shape", mask[1].shape)
        mask = [m[np.newaxis, :] for m in mask]
        mask = np.concatenate(mask, axis=0)
        # print("mask shape after concatenation", mask.shape)
        mask = torch.from_numpy(mask)
        # if len(mask.shape) == 3:
        #     mask = mask.unsqueeze(1)
    elif isinstance(mask[0], torch.Tensor):
        # unseq_flag = True if len(mask) == 1 else False
        mask = torch.cat(mask, dim=0)
        # print("torch mask shape", mask.shape)
        # if unseq_flag:
        #     mask = mask.unsqueeze(0)

    return mask


class GlidePipeline():
    r"""
    Pipeline for GLIDE mask-guided image generation (i.e., editing).

    This model is written from https://github.com/openai/glide-text2im/blob/main/notebooks/inpaint.ipynb
    Parameters:

    """
    def __init__(self, device: str = None):
        # On GPU
        has_cuda = torch.cuda.is_available()
        if device is None:
            device = torch.device('cpu' if not has_cuda else 'cuda')
        else:
            device = torch.device(device)
        # Create base model.
        options = model_and_diffusion_defaults()

        options['inpaint'] = True
        options['use_fp16'] = has_cuda
        options['timestep_respacing'] = '100'  # use 100 diffusion steps for fast sampling
        model, diffusion = create_model_and_diffusion(**options)
        model.eval()
        if has_cuda:
            model.convert_to_fp16()
        model.to(device)
        model.load_state_dict(load_checkpoint('base-inpaint', device))
        # print('total base parameters', sum(x.numel() for x in model.parameters()))
        print("Loaded base model.")

        # Create upsampler model for super-resolution. 64 -> 256
        options_up = model_and_diffusion_defaults_upsampler()
        options_up['inpaint'] = True
        options_up['use_fp16'] = has_cuda
        options_up['timestep_respacing'] = 'fast27'  # use 27 diffusion steps for very fast sampling
        model_up, diffusion_up = create_model_and_diffusion(**options_up)
        model_up.eval()
        if has_cuda:
            model_up.convert_to_fp16()
        model_up.to(device)
        model_up.load_state_dict(load_checkpoint('upsample-inpaint', device))
        # print('total upsampler parameters', sum(x.numel() for x in model_up.parameters()))
        print("Loaded upsampler model.")

        # Save to class.
        self.options = options  # model config
        self.options_up = options_up  # upsampler config
        self.model = model
        self.diffusion = diffusion
        self.model_up = model_up
        self.diffusion_up = diffusion_up
        self.device = device

    def set_seed(self, seed: int):
        torch.manual_seed(seed)
        np.random.seed(seed)

    def batch_tokenization(self, prompts: List[str], tokenizer):
        batch_size = len(prompts)
        batch_tokens, batch_mask = [], []  # token mask
        for prompt in prompts:
            tokens = tokenizer.encode(prompt)
            batch_tokens.append(tokens)

        for i in range(batch_size):
            tokens, mask = tokenizer.padded_tokens_and_mask(
                            batch_tokens[i], self.options['text_ctx'])
            batch_tokens[i] = tokens
            batch_mask.append(mask)

        # Create the classifier-free guidance tokens (empty)
        # full_batch_size = batch_size * 2
        # batch_uncond_tokens, batch_uncond_mask = [], []
        uncond_tokens, uncond_mask = tokenizer.padded_tokens_and_mask(
            [], self.options['text_ctx']
        )
        batch_uncond_tokens = [uncond_tokens] * batch_size
        batch_uncond_mask = [uncond_mask] * batch_size
        return batch_tokens, batch_mask, batch_uncond_tokens, batch_uncond_mask

    # Create an classifier-free guidance sampling function
    def model_fn(self, x_t, ts, **kwargs):
        half = x_t[: len(x_t) // 2]
        combined = torch.cat([half, half], dim=0)
        model_out = self.model(combined, ts, **kwargs)
        eps, rest = model_out[:, :3], model_out[:, 3:]
        cond_eps, uncond_eps = torch.split(eps, len(eps) // 2, dim=0)
        half_eps = uncond_eps + self.guidance_scale * (cond_eps - uncond_eps)
        eps = torch.cat([half_eps, half_eps], dim=0)
        return torch.cat([eps, rest], dim=1)


    @torch.no_grad()
    def __call__(
        self,
        original_image: Union[torch.FloatTensor, PIL.Image.Image],
        mask_image: Union[torch.FloatTensor, PIL.Image.Image],
        prompts: List[str],
        guidance_scale: float = 5.0,
        upsample_temp: float = 0.997,
        # jump_length: int = 10,
        # jump_n_sample: int = 10,
        # generator: Optional[torch.Generator] = None,
        output_type: Optional[str] = "pil",
        # return_dict: bool = True,
    ) -> List[PIL.Image.Image]:


        batch_size = len(prompts)

        self.guidance_scale = guidance_scale

        full_batch_size = batch_size * 2  # for additional classifier free guidance
        batch_tokens, batch_mask, batch_uncond_tokens, batch_uncond_mask = self.batch_tokenization(prompts, self.model.tokenizer)
        # Pack the tokens together into model kwargs.
        original_image_64 = _preprocess_image(original_image, self.options['image_size'])  # [batch, channel, 64, 64]
        mask_image_64 = _preprocess_mask(mask_image, self.options['image_size'])  # [batch, 1, 64, 64]
        original_image_256 = _preprocess_image(original_image, self.options_up['image_size'])  # [batch, channel, 256, 256]
        mask_image_256 = _preprocess_mask(mask_image, self.options_up['image_size'])  # [batch, 1, 256, 256]

        # print("original_image_64", original_image_64.shape, type(original_image_64))
        # print("mask_image_64", mask_image_64.shape, type(mask_image_64))
        # print("original_image_256", original_image_256.shape, type(original_image_256))
        # print("mask_image_256", mask_image_256.shape, type(mask_image_256))

        model_kwargs_base = dict(
            tokens=torch.tensor(
                batch_tokens + batch_uncond_tokens, device=self.device),
            mask=torch.tensor(
                batch_mask + batch_uncond_mask,
                dtype=torch.bool,
                device=self.device,
            ),
            inpaint_image=(original_image_64 * mask_image_64).repeat(2, 1, 1, 1).to(self.device),
            inpaint_mask=mask_image_64.repeat(2, 1, 1, 1).to(self.device),
        )

        print("Sampling...")
        # print("inpaint_image", model_kwargs_base['inpaint_image'].shape)
        # print("inpaint_mask", model_kwargs_base['inpaint_mask'].shape)

        def denoised_fn_base(x_start):
            # Force the model to have the exact right x_start predictions
            # for the part of the image which is known.
            return (
                x_start * (1 - model_kwargs_base['inpaint_mask'])
                + model_kwargs_base['inpaint_image'] * model_kwargs_base['inpaint_mask']
            )

        # Sample from the base model.
        self.model.del_cache()
        samples = self.diffusion.p_sample_loop(
            self.model_fn,
            (full_batch_size, 3, self.options["image_size"], self.options["image_size"]),
            device=self.device,
            clip_denoised=True,
            progress=True,
            model_kwargs=model_kwargs_base,
            cond_fn=None,
            denoised_fn=denoised_fn_base,
        )[:batch_size]
        self.model.del_cache()

        # Upsample the samples.
        batch_tokens, batch_mask, _, _ = self.batch_tokenization(prompts, self.model_up.tokenizer)
        # Create the model conditioning dict.
        model_kwargs_upsample = dict(
            # Low-res image to upsample.
            low_res=((samples+1)*127.5).round()/127.5 - 1,

            # Text tokens
            tokens=torch.tensor(
                batch_tokens, device=self.device
            ),
            mask=torch.tensor(
                batch_mask,
                dtype=torch.bool,
                device=self.device,
            ),

            # Masked inpainting image.
            # inpaint_image=(original_image_256 * mask_image_256).repeat(batch_size, 1, 1, 1).to(self.device),
            # inpaint_mask=mask_image_256.repeat(batch_size, 1, 1, 1).to(self.device),
            inpaint_image=(original_image_256 * mask_image_256).to(self.device),
            inpaint_mask=mask_image_256.to(self.device),
        )

        print("Upsampling...")
        # print("Batch_tokens", batch_tokens)
        # print("Batch_mask", batch_mask)
        # print("Batch_tokens", np.array(batch_tokens).shape)
        # print("Batch_mask", np.array(batch_mask).shape)
        # print("inpaint_image", model_kwargs_upsample['inpaint_image'].shape)
        # print("inpaint_mask", model_kwargs_upsample['inpaint_mask'].shape)

        def denoised_fn_upsample(x_start):
            # Force the model to have the exact right x_start predictions
            # for the part of the image which is known.
            # print("x_start", x_start.shape)
            # print("model_kwargs_upsample['inpaint_mask']", model_kwargs_upsample['inpaint_mask'].shape)
            # print("model_kwargs_upsample['inpaint_image']", model_kwargs_upsample['inpaint_image'].shape)
            return (
                x_start * (1 - model_kwargs_upsample['inpaint_mask'])
                + model_kwargs_upsample['inpaint_image'] * model_kwargs_upsample['inpaint_mask']
            )

        # Sample from the base model.
        self.model_up.del_cache()
        up_shape = (batch_size, 3, self.options_up["image_size"], self.options_up["image_size"])
        up_samples = self.diffusion_up.p_sample_loop(
            self.model_up,
            up_shape,
            noise=torch.randn(up_shape, device=self.device) * upsample_temp,
            device=self.device,
            clip_denoised=True,
            progress=True,
            model_kwargs=model_kwargs_upsample,
            cond_fn=None,
            denoised_fn=denoised_fn_upsample,
        )[:batch_size]
        self.model_up.del_cache()

        scaled_up_samples = ((up_samples + 1)*127.5).round().clamp(0, 255).to(torch.uint8).cpu()
        # transform each image to PIL image
        scaled_up_samples = scaled_up_samples.permute(0, 2, 3, 1).numpy()
        reshaped_up_samples = scaled_up_samples.reshape(batch_size, self.options_up["image_size"], self.options_up["image_size"], 3)

        up_images = [Image.fromarray(reshaped_up_samples[i]) for i in range(batch_size)]
        # up_images = Image.fromarray(reshaped_up_samples.numpy())

        return ImagePipelineOutput(images=up_images)



# modified based on https://github.com/Victarry/stable-dreambooth

from imagen_hub.pipelines.dreambooth.utils.dataset import TrainDataset
from imagen_hub.depend.clip_retrieval.clip_client import ClipClient, Modality

import torch
import torch.nn.functional as F
from accelerate import Accelerator
from diffusers.schedulers import DDPMScheduler
from diffusers.pipelines import StableDiffusionPipeline
from PIL import Image
from tqdm import tqdm
import urllib.request
import logging
import os


class DreamBoothPipeline:
    def __init__(self, device="cuda",
                 subject_name = "dog",
                 identifier = "sks",
                 data_path = None,
                 evaluate_prompt = "photo of a sleeping sks dog",
                 num_epochs: int = 5,
                 train_batch_size: int = 1,
                 learning_rate: float = 1e-5,
                 image_size: int = 512,
                 gradient_accumulation_steps: int = 1,
                 num_train_timesteps: int = 1000,
                 train_guidance_scale: float = 1,
                 sample_guidance_scale: float = 7.5,
                 mixed_precision = 'fp16', save_image_epochs = 1, save_model_epochs = 1,
                 output_dir = None,
                 model_id="CompVis/stable-diffusion-v1-4",
                 num_ref_images=100
                ):
        self.instance_prompt = "photo of a " + identifier + " " + subject_name
        self.class_prompt = "photo of a " + subject_name
        self.evaluate_prompt = evaluate_prompt
        self.train_batch_size = train_batch_size
        self.learning_rate = learning_rate
        self.image_size = image_size
        self.gradient_accumulation_steps = gradient_accumulation_steps
        self.num_train_timesteps = num_train_timesteps
        self.train_guidance_scale = train_guidance_scale
        self.sample_guidance_scale = sample_guidance_scale
        self.mixed_precision = mixed_precision,
        self.save_image_epochs = save_image_epochs,
        self.save_model_epochs = save_model_epochs,
        self.output_dir = output_dir
        self.model_id = model_id
        self.device = device
        self.data_path = data_path
        self.num_epochs = num_epochs
        self.num_ref_images = num_ref_images

    def pred(self, model, noisy_latent, time_steps, prompt, guidance_scale):
        batch_size = noisy_latent.shape[0]
        text_input = model.tokenizer(
            prompt,
            padding="max_length",
            max_length=model.tokenizer.model_max_length,
            truncation=True,
            return_tensors="pt",
        )
        with torch.no_grad():
            text_embeddings = model.text_encoder(text_input.input_ids.to(model.device))[0]
        # here `guidance_scale` is defined analog to the guidance weight `w` of equation (2)
        # of the Imagen paper: https://arxiv.org/pdf/2205.11487.pdf . `guidance_scale = 1`
        # corresponds to doing no classifier free guidance.
        do_classifier_free_guidance = guidance_scale > 1.0
        # get unconditional embeddings for classifier free guidance
        if do_classifier_free_guidance:
            max_length = text_input.input_ids.shape[-1]
            uncond_input = model.tokenizer(
                [""] * batch_size, padding="max_length", max_length=max_length, return_tensors="pt"
            )
            uncond_embeddings = model.text_encoder(uncond_input.input_ids.to(model.device))[0]

            # For classifier free guidance, we need to do two forward passes.
            # Here we concatenate the unconditional and text embeddings into a single batch
            # to avoid doing two forward passes
            text_embeddings = torch.cat([uncond_embeddings, text_embeddings])

        latent_model_input = torch.cat([noisy_latent] * 2) if do_classifier_free_guidance else noisy_latent
        time_steps = torch.cat([time_steps] * 2) if do_classifier_free_guidance else time_steps
        noise_pred = model.unet(latent_model_input, time_steps, encoder_hidden_states=text_embeddings)["sample"]
        # perform guidance
        if do_classifier_free_guidance:
            noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)
            noise_pred = noise_pred_uncond + guidance_scale * (noise_pred_text - noise_pred_uncond)
        return noise_pred

    def retrieve(self):
        # logging_dir = Path(args.output_dir, "0", args.logging_dir)

        logging.basicConfig(
            format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
            datefmt="%m/%d/%Y %H:%M:%S",
            level=logging.INFO,
        )
        logger = logging.getLogger('clip_image_retrieval')

        client = ClipClient(
            url="https://knn.laion.ai/knn-service",
            indice_name="laion5B-H-14",
            aesthetic_score=9,
            aesthetic_weight=0.5,
            modality=Modality.IMAGE,
            num_images=self.num_ref_images)

        results = client.query(text=self.class_prompt)
        logger.info("{} of {} are retrieved".format(len(results), self.class_prompt))
        for i, result in enumerate(results):
            if not result["url"].endswith(".jpg"):
                continue
            if not os.path.exists(os.path.join(self.data_path, "class")):
                os.makedirs(os.path.join(self.data_path, "class"))
            try:
                urllib.request.urlretrieve(result["url"], os.path.join(self.data_path, "class",  "found" + str(i) + ".jpg"))
            except:
                logger.info("failure")
                continue


    def train_loop(self, model: StableDiffusionPipeline, noise_scheduler, optimizer, train_dataloader):
        # Initialize accelerator and tensorboard logging
        print(self.mixed_precision)
        accelerator = Accelerator(
            mixed_precision=self.mixed_precision[0],
            gradient_accumulation_steps=self.gradient_accumulation_steps,

        )
        print(accelerator.device)
        if accelerator.is_main_process:
            accelerator.init_trackers("train_example")

        # Prepare everything
        # There is no specific order to remember, you just need to unpack the
        # objects in the same order you gave them to the prepare method.
        model, optimizer, train_dataloader = accelerator.prepare(
            model, optimizer, train_dataloader
        )

        global_step = 0

        # Now you train the model
        for epoch in range(self.num_epochs):
            progress_bar = tqdm(total=len(train_dataloader), disable=not accelerator.is_local_main_process)
            progress_bar.set_description(f"Epoch {epoch}")

            for step, batch in enumerate(train_dataloader):
                instance_imgs, instance_prompt, class_imgs, class_prompt = batch
                imgs = torch.cat((instance_imgs, class_imgs), dim=0).to(self.device)
                prompt = instance_prompt + class_prompt

                # Sample noise to add to the images
                bs = imgs.shape[0]

                # Sample a random timestep for each image
                timesteps = torch.randint(0, noise_scheduler.num_train_timesteps, (bs,), device=self.device).long()

                # Add noise to the clean images according to the noise magnitude at each timestep
                # (this is the forward diffusion process)
                with torch.no_grad():
                    latents = model.vae.encode(imgs).latent_dist.sample() * 0.18215
                    noise = torch.randn(latents.shape, device=self.device)
                    # noisy_latents = noise_scheduler.add_noise(latents, noise, timesteps.cpu().numpy())
                    noisy_latents = noise_scheduler.add_noise(latents, noise, timesteps)

                with accelerator.accumulate(model):
                    # Predict the noise residual
                    noise_pred = self.pred(model, noisy_latents, timesteps, prompt, guidance_scale=self.train_guidance_scale)
                    loss = F.mse_loss(noise_pred, noise)
                    accelerator.backward(loss)

                    accelerator.clip_grad_norm_(model.unet.parameters(), 1.0)
                    optimizer.step()
                    optimizer.zero_grad()

                progress_bar.update(1)
                logs = {"loss": loss.detach().item(), "step": global_step}
                progress_bar.set_postfix(**logs)
                accelerator.log(logs, step=global_step)
                global_step += 1

            # After each epoch you optionally sample some demo images with evaluate() and save the model
            if accelerator.is_main_process:
                # if epoch % self.save_image_epochs[0] == 0 or epoch == self.num_epochs - 1:
                #     self.evaluate(self.evaluate_prompt, epoch, model)

                if (epoch + 1) % self.save_model_epochs[0] == 0 or epoch == self.num_epochs - 1:
                    model.save_pretrained(self.output_dir)


    def make_grid(self, images, rows, cols):
        w, h = images[0].size
        grid = Image.new('RGB', size=(cols * w, rows * h))
        for i, image in enumerate(images):
            grid.paste(image, box=(i % cols * w, i // cols * h))
        return grid


    def evaluate(self, evaluate_prompt, epoch, pipeline: StableDiffusionPipeline):
        # Sample some images from random noise (this is the backward diffusion process).
        # The default pipeline output type is `List[PIL.Image]`
        with torch.no_grad():
            # with torch.autocast("cuda"):
            pipeline.to("cuda")
            images = \
            pipeline(evaluate_prompt, num_inference_steps=50, width=self.image_size, height=self.image_size,
                     guidance_scale=self.sample_guidance_scale)
        print(images)
        print(images[0])
        print(len(images))

        # Make a grid out of the images
        image_grid = self.make_grid(images[0], rows=4, cols=4)

        # Save the images
        test_dir = os.path.join(self.output_dir, "samples")
        os.makedirs(test_dir, exist_ok=True)
        image_grid.save(f"{test_dir}/{epoch:04d}.jpg")

    def evaluate_one(self, evaluate_prompt, pipeline: StableDiffusionPipeline, seed: int):
        # Sample some images from random noise (this is the backward diffusion process).
        # The default pipeline output type is `List[PIL.Image]`
        with torch.no_grad():
            with torch.autocast("cuda"):
                generator = [torch.Generator(device="cuda").manual_seed(seed)]
                images = \
                pipeline([evaluate_prompt], num_inference_steps=50, generator=generator, width=self.image_size, height=self.image_size,
                         guidance_scale=self.sample_guidance_scale)

        return images[0][0]




    def get_dataloader(self,):
        dataset = TrainDataset(self.data_path, self.instance_prompt, self.class_prompt, self.image_size)
        dataloader = torch.utils.data.DataLoader(dataset, batch_size=self.train_batch_size, shuffle=True, drop_last=True,
                                                 pin_memory=True)
        return dataloader

    def train(self,):
        self.retrieve()
        train_dataloader = self.get_dataloader()
        model = StableDiffusionPipeline.from_pretrained(self.model_id, use_auth_token=True,
                                                        cache_dir="./.cache").to(self.device)
        optimizer = torch.optim.AdamW(model.unet.parameters(), lr=self.learning_rate)
        noise_scheduler = DDPMScheduler(num_train_timesteps=self.num_train_timesteps, beta_start=0.00085,
                                        beta_end=0.0120)
        self.train_loop(model, noise_scheduler, optimizer, train_dataloader)



import torch.nn as nn
from mmcv.cnn import ConvModule, normal_init
from mmcv.ops import DeformConv2d

from mmdet.core import multi_apply
from ..builder import HEADS, build_loss
from .corner_head import CornerHead


@HEADS.register_module()
class CentripetalHead(CornerHead):
    """Head of CentripetalNet: Pursuing High-quality Keypoint Pairs for Object
    Detection.

    CentripetalHead inherits from :class:`CornerHead`. It removes the
    embedding branch and adds guiding shift and centripetal shift branches.
    More details can be found in the `paper
    <https://arxiv.org/abs/2003.09119>`_ .

    Args:
        num_classes (int): Number of categories excluding the background
            category.
        in_channels (int): Number of channels in the input feature map.
        num_feat_levels (int): Levels of feature from the previous module. 2
            for HourglassNet-104 and 1 for HourglassNet-52. HourglassNet-104
            outputs the final feature and intermediate supervision feature and
            HourglassNet-52 only outputs the final feature. Default: 2.
        corner_emb_channels (int): Channel of embedding vector. Default: 1.
        train_cfg (dict | None): Training config. Useless in CornerHead,
            but we keep this variable for SingleStageDetector. Default: None.
        test_cfg (dict | None): Testing config of CornerHead. Default: None.
        loss_heatmap (dict | None): Config of corner heatmap loss. Default:
            GaussianFocalLoss.
        loss_embedding (dict | None): Config of corner embedding loss. Default:
            AssociativeEmbeddingLoss.
        loss_offset (dict | None): Config of corner offset loss. Default:
            SmoothL1Loss.
        loss_guiding_shift (dict): Config of guiding shift loss. Default:
            SmoothL1Loss.
        loss_centripetal_shift (dict): Config of centripetal shift loss.
            Default: SmoothL1Loss.
    """

    def __init__(self,
                 *args,
                 centripetal_shift_channels=2,
                 guiding_shift_channels=2,
                 feat_adaption_conv_kernel=3,
                 loss_guiding_shift=dict(
                     type='SmoothL1Loss', beta=1.0, loss_weight=0.05),
                 loss_centripetal_shift=dict(
                     type='SmoothL1Loss', beta=1.0, loss_weight=1),
                 **kwargs):
        assert centripetal_shift_channels == 2, (
            'CentripetalHead only support centripetal_shift_channels == 2')
        self.centripetal_shift_channels = centripetal_shift_channels
        assert guiding_shift_channels == 2, (
            'CentripetalHead only support guiding_shift_channels == 2')
        self.guiding_shift_channels = guiding_shift_channels
        self.feat_adaption_conv_kernel = feat_adaption_conv_kernel
        super(CentripetalHead, self).__init__(*args, **kwargs)
        self.loss_guiding_shift = build_loss(loss_guiding_shift)
        self.loss_centripetal_shift = build_loss(loss_centripetal_shift)

    def _init_centripetal_layers(self):
        """Initialize centripetal layers.

        Including feature adaption deform convs (feat_adaption), deform offset
        prediction convs (dcn_off), guiding shift (guiding_shift) and
        centripetal shift ( centripetal_shift). Each branch has two parts:
        prefix `tl_` for top-left and `br_` for bottom-right.
        """
        self.tl_feat_adaption = nn.ModuleList()
        self.br_feat_adaption = nn.ModuleList()
        self.tl_dcn_offset = nn.ModuleList()
        self.br_dcn_offset = nn.ModuleList()
        self.tl_guiding_shift = nn.ModuleList()
        self.br_guiding_shift = nn.ModuleList()
        self.tl_centripetal_shift = nn.ModuleList()
        self.br_centripetal_shift = nn.ModuleList()

        for _ in range(self.num_feat_levels):
            self.tl_feat_adaption.append(
                DeformConv2d(self.in_channels, self.in_channels,
                             self.feat_adaption_conv_kernel, 1, 1))
            self.br_feat_adaption.append(
                DeformConv2d(self.in_channels, self.in_channels,
                             self.feat_adaption_conv_kernel, 1, 1))

            self.tl_guiding_shift.append(
                self._make_layers(
                    out_channels=self.guiding_shift_channels,
                    in_channels=self.in_channels))
            self.br_guiding_shift.append(
                self._make_layers(
                    out_channels=self.guiding_shift_channels,
                    in_channels=self.in_channels))

            self.tl_dcn_offset.append(
                ConvModule(
                    self.guiding_shift_channels,
                    self.feat_adaption_conv_kernel**2 *
                    self.guiding_shift_channels,
                    1,
                    bias=False,
                    act_cfg=None))
            self.br_dcn_offset.append(
                ConvModule(
                    self.guiding_shift_channels,
                    self.feat_adaption_conv_kernel**2 *
                    self.guiding_shift_channels,
                    1,
                    bias=False,
                    act_cfg=None))

            self.tl_centripetal_shift.append(
                self._make_layers(
                    out_channels=self.centripetal_shift_channels,
                    in_channels=self.in_channels))
            self.br_centripetal_shift.append(
                self._make_layers(
                    out_channels=self.centripetal_shift_channels,
                    in_channels=self.in_channels))

    def _init_layers(self):
        """Initialize layers for CentripetalHead.

        Including two parts: CornerHead layers and CentripetalHead layers
        """
        super()._init_layers()  # using _init_layers in CornerHead
        self._init_centripetal_layers()

    def init_weights(self):
        """Initialize weights of the head."""
        super().init_weights()
        for i in range(self.num_feat_levels):
            normal_init(self.tl_feat_adaption[i], std=0.01)
            normal_init(self.br_feat_adaption[i], std=0.01)
            normal_init(self.tl_dcn_offset[i].conv, std=0.1)
            normal_init(self.br_dcn_offset[i].conv, std=0.1)
            _ = [x.conv.reset_parameters() for x in self.tl_guiding_shift[i]]
            _ = [x.conv.reset_parameters() for x in self.br_guiding_shift[i]]
            _ = [
                x.conv.reset_parameters() for x in self.tl_centripetal_shift[i]
            ]
            _ = [
                x.conv.reset_parameters() for x in self.br_centripetal_shift[i]
            ]

    def forward_single(self, x, lvl_ind):
        """Forward feature of a single level.

        Args:
            x (Tensor): Feature of a single level.
            lvl_ind (int): Level index of current feature.

        Returns:
            tuple[Tensor]: A tuple of CentripetalHead's output for current
            feature level. Containing the following Tensors:

                - tl_heat (Tensor): Predicted top-left corner heatmap.
                - br_heat (Tensor): Predicted bottom-right corner heatmap.
                - tl_off (Tensor): Predicted top-left offset heatmap.
                - br_off (Tensor): Predicted bottom-right offset heatmap.
                - tl_guiding_shift (Tensor): Predicted top-left guiding shift
                  heatmap.
                - br_guiding_shift (Tensor): Predicted bottom-right guiding
                  shift heatmap.
                - tl_centripetal_shift (Tensor): Predicted top-left centripetal
                  shift heatmap.
                - br_centripetal_shift (Tensor): Predicted bottom-right
                  centripetal shift heatmap.
        """
        tl_heat, br_heat, _, _, tl_off, br_off, tl_pool, br_pool = super(
        ).forward_single(
            x, lvl_ind, return_pool=True)

        tl_guiding_shift = self.tl_guiding_shift[lvl_ind](tl_pool)
        br_guiding_shift = self.br_guiding_shift[lvl_ind](br_pool)

        tl_dcn_offset = self.tl_dcn_offset[lvl_ind](tl_guiding_shift.detach())
        br_dcn_offset = self.br_dcn_offset[lvl_ind](br_guiding_shift.detach())

        tl_feat_adaption = self.tl_feat_adaption[lvl_ind](tl_pool,
                                                          tl_dcn_offset)
        br_feat_adaption = self.br_feat_adaption[lvl_ind](br_pool,
                                                          br_dcn_offset)

        tl_centripetal_shift = self.tl_centripetal_shift[lvl_ind](
            tl_feat_adaption)
        br_centripetal_shift = self.br_centripetal_shift[lvl_ind](
            br_feat_adaption)

        result_list = [
            tl_heat, br_heat, tl_off, br_off, tl_guiding_shift,
            br_guiding_shift, tl_centripetal_shift, br_centripetal_shift
        ]
        return result_list

    def loss(self,
             tl_heats,
             br_heats,
             tl_offs,
             br_offs,
             tl_guiding_shifts,
             br_guiding_shifts,
             tl_centripetal_shifts,
             br_centripetal_shifts,
             gt_bboxes,
             gt_labels,
             img_metas,
             gt_bboxes_ignore=None):
        """Compute losses of the head.

        Args:
            tl_heats (list[Tensor]): Top-left corner heatmaps for each level
                with shape (N, num_classes, H, W).
            br_heats (list[Tensor]): Bottom-right corner heatmaps for each
                level with shape (N, num_classes, H, W).
            tl_offs (list[Tensor]): Top-left corner offsets for each level
                with shape (N, corner_offset_channels, H, W).
            br_offs (list[Tensor]): Bottom-right corner offsets for each level
                with shape (N, corner_offset_channels, H, W).
            tl_guiding_shifts (list[Tensor]): Top-left guiding shifts for each
                level with shape (N, guiding_shift_channels, H, W).
            br_guiding_shifts (list[Tensor]): Bottom-right guiding shifts for
                each level with shape (N, guiding_shift_channels, H, W).
            tl_centripetal_shifts (list[Tensor]): Top-left centripetal shifts
                for each level with shape (N, centripetal_shift_channels, H,
                W).
            br_centripetal_shifts (list[Tensor]): Bottom-right centripetal
                shifts for each level with shape (N,
                centripetal_shift_channels, H, W).
            gt_bboxes (list[Tensor]): Ground truth bboxes for each image with
                shape (num_gts, 4) in [left, top, right, bottom] format.
            gt_labels (list[Tensor]): Class indices corresponding to each box.
            img_metas (list[dict]): Meta information of each image, e.g.,
                image size, scaling factor, etc.
            gt_bboxes_ignore (list[Tensor] | None): Specify which bounding
                boxes can be ignored when computing the loss.

        Returns:
            dict[str, Tensor]: A dictionary of loss components. Containing the
            following losses:

                - det_loss (list[Tensor]): Corner keypoint losses of all
                  feature levels.
                - off_loss (list[Tensor]): Corner offset losses of all feature
                  levels.
                - guiding_loss (list[Tensor]): Guiding shift losses of all
                  feature levels.
                - centripetal_loss (list[Tensor]): Centripetal shift losses of
                  all feature levels.
        """
        targets = self.get_targets(
            gt_bboxes,
            gt_labels,
            tl_heats[-1].shape,
            img_metas[0]['pad_shape'],
            with_corner_emb=self.with_corner_emb,
            with_guiding_shift=True,
            with_centripetal_shift=True)
        mlvl_targets = [targets for _ in range(self.num_feat_levels)]
        [det_losses, off_losses, guiding_losses, centripetal_losses
         ] = multi_apply(self.loss_single, tl_heats, br_heats, tl_offs,
                         br_offs, tl_guiding_shifts, br_guiding_shifts,
                         tl_centripetal_shifts, br_centripetal_shifts,
                         mlvl_targets)
        loss_dict = dict(
            det_loss=det_losses,
            off_loss=off_losses,
            guiding_loss=guiding_losses,
            centripetal_loss=centripetal_losses)
        return loss_dict

    def loss_single(self, tl_hmp, br_hmp, tl_off, br_off, tl_guiding_shift,
                    br_guiding_shift, tl_centripetal_shift,
                    br_centripetal_shift, targets):
        """Compute losses for single level.

        Args:
            tl_hmp (Tensor): Top-left corner heatmap for current level with
                shape (N, num_classes, H, W).
            br_hmp (Tensor): Bottom-right corner heatmap for current level with
                shape (N, num_classes, H, W).
            tl_off (Tensor): Top-left corner offset for current level with
                shape (N, corner_offset_channels, H, W).
            br_off (Tensor): Bottom-right corner offset for current level with
                shape (N, corner_offset_channels, H, W).
            tl_guiding_shift (Tensor): Top-left guiding shift for current level
                with shape (N, guiding_shift_channels, H, W).
            br_guiding_shift (Tensor): Bottom-right guiding shift for current
                level with shape (N, guiding_shift_channels, H, W).
            tl_centripetal_shift (Tensor): Top-left centripetal shift for
                current level with shape (N, centripetal_shift_channels, H, W).
            br_centripetal_shift (Tensor): Bottom-right centripetal shift for
                current level with shape (N, centripetal_shift_channels, H, W).
            targets (dict): Corner target generated by `get_targets`.

        Returns:
            tuple[torch.Tensor]: Losses of the head's differnet branches
            containing the following losses:

                - det_loss (Tensor): Corner keypoint loss.
                - off_loss (Tensor): Corner offset loss.
                - guiding_loss (Tensor): Guiding shift loss.
                - centripetal_loss (Tensor): Centripetal shift loss.
        """
        targets['corner_embedding'] = None

        det_loss, _, _, off_loss = super().loss_single(tl_hmp, br_hmp, None,
                                                       None, tl_off, br_off,
                                                       targets)

        gt_tl_guiding_shift = targets['topleft_guiding_shift']
        gt_br_guiding_shift = targets['bottomright_guiding_shift']
        gt_tl_centripetal_shift = targets['topleft_centripetal_shift']
        gt_br_centripetal_shift = targets['bottomright_centripetal_shift']

        gt_tl_heatmap = targets['topleft_heatmap']
        gt_br_heatmap = targets['bottomright_heatmap']
        # We only compute the offset loss at the real corner position.
        # The value of real corner would be 1 in heatmap ground truth.
        # The mask is computed in class agnostic mode and its shape is
        # batch * 1 * width * height.
        tl_mask = gt_tl_heatmap.eq(1).sum(1).gt(0).unsqueeze(1).type_as(
            gt_tl_heatmap)
        br_mask = gt_br_heatmap.eq(1).sum(1).gt(0).unsqueeze(1).type_as(
            gt_br_heatmap)

        # Guiding shift loss
        tl_guiding_loss = self.loss_guiding_shift(
            tl_guiding_shift,
            gt_tl_guiding_shift,
            tl_mask,
            avg_factor=tl_mask.sum())
        br_guiding_loss = self.loss_guiding_shift(
            br_guiding_shift,
            gt_br_guiding_shift,
            br_mask,
            avg_factor=br_mask.sum())
        guiding_loss = (tl_guiding_loss + br_guiding_loss) / 2.0
        # Centripetal shift loss
        tl_centripetal_loss = self.loss_centripetal_shift(
            tl_centripetal_shift,
            gt_tl_centripetal_shift,
            tl_mask,
            avg_factor=tl_mask.sum())
        br_centripetal_loss = self.loss_centripetal_shift(
            br_centripetal_shift,
            gt_br_centripetal_shift,
            br_mask,
            avg_factor=br_mask.sum())
        centripetal_loss = (tl_centripetal_loss + br_centripetal_loss) / 2.0

        return det_loss, off_loss, guiding_loss, centripetal_loss

    def get_bboxes(self,
                   tl_heats,
                   br_heats,
                   tl_offs,
                   br_offs,
                   tl_guiding_shifts,
                   br_guiding_shifts,
                   tl_centripetal_shifts,
                   br_centripetal_shifts,
                   img_metas,
                   rescale=False,
                   with_nms=True):
        """Transform network output for a batch into bbox predictions.

        Args:
            tl_heats (list[Tensor]): Top-left corner heatmaps for each level
                with shape (N, num_classes, H, W).
            br_heats (list[Tensor]): Bottom-right corner heatmaps for each
                level with shape (N, num_classes, H, W).
            tl_offs (list[Tensor]): Top-left corner offsets for each level
                with shape (N, corner_offset_channels, H, W).
            br_offs (list[Tensor]): Bottom-right corner offsets for each level
                with shape (N, corner_offset_channels, H, W).
            tl_guiding_shifts (list[Tensor]): Top-left guiding shifts for each
                level with shape (N, guiding_shift_channels, H, W). Useless in
                this function, we keep this arg because it's the raw output
                from CentripetalHead.
            br_guiding_shifts (list[Tensor]): Bottom-right guiding shifts for
                each level with shape (N, guiding_shift_channels, H, W).
                Useless in this function, we keep this arg because it's the
                raw output from CentripetalHead.
            tl_centripetal_shifts (list[Tensor]): Top-left centripetal shifts
                for each level with shape (N, centripetal_shift_channels, H,
                W).
            br_centripetal_shifts (list[Tensor]): Bottom-right centripetal
                shifts for each level with shape (N,
                centripetal_shift_channels, H, W).
            img_metas (list[dict]): Meta information of each image, e.g.,
                image size, scaling factor, etc.
            rescale (bool): If True, return boxes in original image space.
                Default: False.
            with_nms (bool): If True, do nms before return boxes.
                Default: True.
        """
        assert tl_heats[-1].shape[0] == br_heats[-1].shape[0] == len(img_metas)
        result_list = []
        for img_id in range(len(img_metas)):
            result_list.append(
                self._get_bboxes_single(
                    tl_heats[-1][img_id:img_id + 1, :],
                    br_heats[-1][img_id:img_id + 1, :],
                    tl_offs[-1][img_id:img_id + 1, :],
                    br_offs[-1][img_id:img_id + 1, :],
                    img_metas[img_id],
                    tl_emb=None,
                    br_emb=None,
                    tl_centripetal_shift=tl_centripetal_shifts[-1][
                        img_id:img_id + 1, :],
                    br_centripetal_shift=br_centripetal_shifts[-1][
                        img_id:img_id + 1, :],
                    rescale=rescale,
                    with_nms=with_nms))

        return result_list



import numpy as np

from ..builder import PIPELINES


@PIPELINES.register_module()
class InstaBoost(object):
    r"""Data augmentation method in `InstaBoost: Boosting Instance
    Segmentation Via Probability Map Guided Copy-Pasting
    <https://arxiv.org/abs/1908.07801>`_.

    Refer to https://github.com/GothicAi/Instaboost for implementation details.
    """

    def __init__(self,
                 action_candidate=('normal', 'horizontal', 'skip'),
                 action_prob=(1, 0, 0),
                 scale=(0.8, 1.2),
                 dx=15,
                 dy=15,
                 theta=(-1, 1),
                 color_prob=0.5,
                 hflag=False,
                 aug_ratio=0.5):
        try:
            import instaboostfast as instaboost
        except ImportError:
            raise ImportError(
                'Please run "pip install instaboostfast" '
                'to install instaboostfast first for instaboost augmentation.')
        self.cfg = instaboost.InstaBoostConfig(action_candidate, action_prob,
                                               scale, dx, dy, theta,
                                               color_prob, hflag)
        self.aug_ratio = aug_ratio

    def _load_anns(self, results):
        labels = results['ann_info']['labels']
        masks = results['ann_info']['masks']
        bboxes = results['ann_info']['bboxes']
        n = len(labels)

        anns = []
        for i in range(n):
            label = labels[i]
            bbox = bboxes[i]
            mask = masks[i]
            x1, y1, x2, y2 = bbox
            # assert (x2 - x1) >= 1 and (y2 - y1) >= 1
            bbox = [x1, y1, x2 - x1, y2 - y1]
            anns.append({
                'category_id': label,
                'segmentation': mask,
                'bbox': bbox
            })

        return anns

    def _parse_anns(self, results, anns, img):
        gt_bboxes = []
        gt_labels = []
        gt_masks_ann = []
        for ann in anns:
            x1, y1, w, h = ann['bbox']
            # TODO: more essential bug need to be fixed in instaboost
            if w <= 0 or h <= 0:
                continue
            bbox = [x1, y1, x1 + w, y1 + h]
            gt_bboxes.append(bbox)
            gt_labels.append(ann['category_id'])
            gt_masks_ann.append(ann['segmentation'])
        gt_bboxes = np.array(gt_bboxes, dtype=np.float32)
        gt_labels = np.array(gt_labels, dtype=np.int64)
        results['ann_info']['labels'] = gt_labels
        results['ann_info']['bboxes'] = gt_bboxes
        results['ann_info']['masks'] = gt_masks_ann
        results['img'] = img
        return results

    def __call__(self, results):
        img = results['img']
        orig_type = img.dtype
        anns = self._load_anns(results)
        if np.random.choice([0, 1], p=[1 - self.aug_ratio, self.aug_ratio]):
            try:
                import instaboostfast as instaboost
            except ImportError:
                raise ImportError('Please run "pip install instaboostfast" '
                                  'to install instaboostfast first.')
            anns, img = instaboost.get_new_data(
                anns, img.astype(np.uint8), self.cfg, background=None)

        results = self._parse_anns(results, anns, img.astype(orig_type))
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f'(cfg={self.cfg}, aug_ratio={self.aug_ratio})'
        return repr_str



"""This module defines the :class:`NiceRepr` mixin class, which defines a
``__repr__`` and ``__str__`` method that only depend on a custom ``__nice__``
method, which you must define. This means you only have to overload one
function instead of two.  Furthermore, if the object defines a ``__len__``
method, then the ``__nice__`` method defaults to something sensible, otherwise
it is treated as abstract and raises ``NotImplementedError``.

To use simply have your object inherit from :class:`NiceRepr`
(multi-inheritance should be ok).

This code was copied from the ubelt library: https://github.com/Erotemic/ubelt

Example:
    >>> # Objects that define __nice__ have a default __str__ and __repr__
    >>> class Student(NiceRepr):
    ...    def __init__(self, name):
    ...        self.name = name
    ...    def __nice__(self):
    ...        return self.name
    >>> s1 = Student('Alice')
    >>> s2 = Student('Bob')
    >>> print(f's1 = {s1}')
    >>> print(f's2 = {s2}')
    s1 = <Student(Alice)>
    s2 = <Student(Bob)>

Example:
    >>> # Objects that define __len__ have a default __nice__
    >>> class Group(NiceRepr):
    ...    def __init__(self, data):
    ...        self.data = data
    ...    def __len__(self):
    ...        return len(self.data)
    >>> g = Group([1, 2, 3])
    >>> print(f'g = {g}')
    g = <Group(3)>
"""
import warnings


class NiceRepr(object):
    """Inherit from this class and define ``__nice__`` to "nicely" print your
    objects.

    Defines ``__str__`` and ``__repr__`` in terms of ``__nice__`` function
    Classes that inherit from :class:`NiceRepr` should redefine ``__nice__``.
    If the inheriting class has a ``__len__``, method then the default
    ``__nice__`` method will return its length.

    Example:
        >>> class Foo(NiceRepr):
        ...    def __nice__(self):
        ...        return 'info'
        >>> foo = Foo()
        >>> assert str(foo) == '<Foo(info)>'
        >>> assert repr(foo).startswith('<Foo(info) at ')

    Example:
        >>> class Bar(NiceRepr):
        ...    pass
        >>> bar = Bar()
        >>> import pytest
        >>> with pytest.warns(None) as record:
        >>>     assert 'object at' in str(bar)
        >>>     assert 'object at' in repr(bar)

    Example:
        >>> class Baz(NiceRepr):
        ...    def __len__(self):
        ...        return 5
        >>> baz = Baz()
        >>> assert str(baz) == '<Baz(5)>'
    """

    def __nice__(self):
        """str: a "nice" summary string describing this module"""
        if hasattr(self, '__len__'):
            # It is a common pattern for objects to use __len__ in __nice__
            # As a convenience we define a default __nice__ for these objects
            return str(len(self))
        else:
            # In all other cases force the subclass to overload __nice__
            raise NotImplementedError(
                f'Define the __nice__ method for {self.__class__!r}')

    def __repr__(self):
        """str: the string of the module"""
        try:
            nice = self.__nice__()
            classname = self.__class__.__name__
            return f'<{classname}({nice}) at {hex(id(self))}>'
        except NotImplementedError as ex:
            warnings.warn(str(ex), category=RuntimeWarning)
            return object.__repr__(self)

    def __str__(self):
        """str: the string of the module"""
        try:
            classname = self.__class__.__name__
            nice = self.__nice__()
            return f'<{classname}({nice})>'
        except NotImplementedError as ex:
            warnings.warn(str(ex), category=RuntimeWarning)
            return object.__repr__(self)



import annotator.uniformer.mmcv as mmcv
import numpy as np
import torch

from ..builder import BBOX_CODERS
from .base_bbox_coder import BaseBBoxCoder


@BBOX_CODERS.register_module()
class DeltaXYWHBBoxCoder(BaseBBoxCoder):
    """Delta XYWH BBox coder.

    Following the practice in `R-CNN <https://arxiv.org/abs/1311.2524>`_,
    this coder encodes bbox (x1, y1, x2, y2) into delta (dx, dy, dw, dh) and
    decodes delta (dx, dy, dw, dh) back to original bbox (x1, y1, x2, y2).

    Args:
        target_means (Sequence[float]): Denormalizing means of target for
            delta coordinates
        target_stds (Sequence[float]): Denormalizing standard deviation of
            target for delta coordinates
        clip_border (bool, optional): Whether clip the objects outside the
            border of the image. Defaults to True.
    """

    def __init__(self,
                 target_means=(0., 0., 0., 0.),
                 target_stds=(1., 1., 1., 1.),
                 clip_border=True):
        super(BaseBBoxCoder, self).__init__()
        self.means = target_means
        self.stds = target_stds
        self.clip_border = clip_border

    def encode(self, bboxes, gt_bboxes):
        """Get box regression transformation deltas that can be used to
        transform the ``bboxes`` into the ``gt_bboxes``.

        Args:
            bboxes (torch.Tensor): Source boxes, e.g., object proposals.
            gt_bboxes (torch.Tensor): Target of the transformation, e.g.,
                ground-truth boxes.

        Returns:
            torch.Tensor: Box transformation deltas
        """

        assert bboxes.size(0) == gt_bboxes.size(0)
        assert bboxes.size(-1) == gt_bboxes.size(-1) == 4
        encoded_bboxes = bbox2delta(bboxes, gt_bboxes, self.means, self.stds)
        return encoded_bboxes

    def decode(self,
               bboxes,
               pred_bboxes,
               max_shape=None,
               wh_ratio_clip=16 / 1000):
        """Apply transformation `pred_bboxes` to `boxes`.

        Args:
            bboxes (torch.Tensor): Basic boxes. Shape (B, N, 4) or (N, 4)
            pred_bboxes (Tensor): Encoded offsets with respect to each roi.
               Has shape (B, N, num_classes * 4) or (B, N, 4) or
               (N, num_classes * 4) or (N, 4). Note N = num_anchors * W * H
               when rois is a grid of anchors.Offset encoding follows [1]_.
            max_shape (Sequence[int] or torch.Tensor or Sequence[
               Sequence[int]],optional): Maximum bounds for boxes, specifies
               (H, W, C) or (H, W). If bboxes shape is (B, N, 4), then
               the max_shape should be a Sequence[Sequence[int]]
               and the length of max_shape should also be B.
            wh_ratio_clip (float, optional): The allowed ratio between
                width and height.

        Returns:
            torch.Tensor: Decoded boxes.
        """

        assert pred_bboxes.size(0) == bboxes.size(0)
        if pred_bboxes.ndim == 3:
            assert pred_bboxes.size(1) == bboxes.size(1)
        decoded_bboxes = delta2bbox(bboxes, pred_bboxes, self.means, self.stds,
                                    max_shape, wh_ratio_clip, self.clip_border)

        return decoded_bboxes


@mmcv.jit(coderize=True)
def bbox2delta(proposals, gt, means=(0., 0., 0., 0.), stds=(1., 1., 1., 1.)):
    """Compute deltas of proposals w.r.t. gt.

    We usually compute the deltas of x, y, w, h of proposals w.r.t ground
    truth bboxes to get regression target.
    This is the inverse function of :func:`delta2bbox`.

    Args:
        proposals (Tensor): Boxes to be transformed, shape (N, ..., 4)
        gt (Tensor): Gt bboxes to be used as base, shape (N, ..., 4)
        means (Sequence[float]): Denormalizing means for delta coordinates
        stds (Sequence[float]): Denormalizing standard deviation for delta
            coordinates

    Returns:
        Tensor: deltas with shape (N, 4), where columns represent dx, dy,
            dw, dh.
    """
    assert proposals.size() == gt.size()

    proposals = proposals.float()
    gt = gt.float()
    px = (proposals[..., 0] + proposals[..., 2]) * 0.5
    py = (proposals[..., 1] + proposals[..., 3]) * 0.5
    pw = proposals[..., 2] - proposals[..., 0]
    ph = proposals[..., 3] - proposals[..., 1]

    gx = (gt[..., 0] + gt[..., 2]) * 0.5
    gy = (gt[..., 1] + gt[..., 3]) * 0.5
    gw = gt[..., 2] - gt[..., 0]
    gh = gt[..., 3] - gt[..., 1]

    dx = (gx - px) / pw
    dy = (gy - py) / ph
    dw = torch.log(gw / pw)
    dh = torch.log(gh / ph)
    deltas = torch.stack([dx, dy, dw, dh], dim=-1)

    means = deltas.new_tensor(means).unsqueeze(0)
    stds = deltas.new_tensor(stds).unsqueeze(0)
    deltas = deltas.sub_(means).div_(stds)

    return deltas


@mmcv.jit(coderize=True)
def delta2bbox(rois,
               deltas,
               means=(0., 0., 0., 0.),
               stds=(1., 1., 1., 1.),
               max_shape=None,
               wh_ratio_clip=16 / 1000,
               clip_border=True):
    """Apply deltas to shift/scale base boxes.

    Typically the rois are anchor or proposed bounding boxes and the deltas are
    network outputs used to shift/scale those boxes.
    This is the inverse function of :func:`bbox2delta`.

    Args:
        rois (Tensor): Boxes to be transformed. Has shape (N, 4) or (B, N, 4)
        deltas (Tensor): Encoded offsets with respect to each roi.
            Has shape (B, N, num_classes * 4) or (B, N, 4) or
            (N, num_classes * 4) or (N, 4). Note N = num_anchors * W * H
            when rois is a grid of anchors.Offset encoding follows [1]_.
        means (Sequence[float]): Denormalizing means for delta coordinates
        stds (Sequence[float]): Denormalizing standard deviation for delta
            coordinates
        max_shape (Sequence[int] or torch.Tensor or Sequence[
            Sequence[int]],optional): Maximum bounds for boxes, specifies
            (H, W, C) or (H, W). If rois shape is (B, N, 4), then
            the max_shape should be a Sequence[Sequence[int]]
            and the length of max_shape should also be B.
        wh_ratio_clip (float): Maximum aspect ratio for boxes.
        clip_border (bool, optional): Whether clip the objects outside the
            border of the image. Defaults to True.

    Returns:
        Tensor: Boxes with shape (B, N, num_classes * 4) or (B, N, 4) or
           (N, num_classes * 4) or (N, 4), where 4 represent
           tl_x, tl_y, br_x, br_y.

    References:
        .. [1] https://arxiv.org/abs/1311.2524

    Example:
        >>> rois = torch.Tensor([[ 0.,  0.,  1.,  1.],
        >>>                      [ 0.,  0.,  1.,  1.],
        >>>                      [ 0.,  0.,  1.,  1.],
        >>>                      [ 5.,  5.,  5.,  5.]])
        >>> deltas = torch.Tensor([[  0.,   0.,   0.,   0.],
        >>>                        [  1.,   1.,   1.,   1.],
        >>>                        [  0.,   0.,   2.,  -1.],
        >>>                        [ 0.7, -1.9, -0.5,  0.3]])
        >>> delta2bbox(rois, deltas, max_shape=(32, 32, 3))
        tensor([[0.0000, 0.0000, 1.0000, 1.0000],
                [0.1409, 0.1409, 2.8591, 2.8591],
                [0.0000, 0.3161, 4.1945, 0.6839],
                [5.0000, 5.0000, 5.0000, 5.0000]])
    """
    means = deltas.new_tensor(means).view(1,
                                          -1).repeat(1,
                                                     deltas.size(-1) // 4)
    stds = deltas.new_tensor(stds).view(1, -1).repeat(1, deltas.size(-1) // 4)
    denorm_deltas = deltas * stds + means
    dx = denorm_deltas[..., 0::4]
    dy = denorm_deltas[..., 1::4]
    dw = denorm_deltas[..., 2::4]
    dh = denorm_deltas[..., 3::4]
    max_ratio = np.abs(np.log(wh_ratio_clip))
    dw = dw.clamp(min=-max_ratio, max=max_ratio)
    dh = dh.clamp(min=-max_ratio, max=max_ratio)
    x1, y1 = rois[..., 0], rois[..., 1]
    x2, y2 = rois[..., 2], rois[..., 3]
    # Compute center of each roi
    px = ((x1 + x2) * 0.5).unsqueeze(-1).expand_as(dx)
    py = ((y1 + y2) * 0.5).unsqueeze(-1).expand_as(dy)
    # Compute width/height of each roi
    pw = (x2 - x1).unsqueeze(-1).expand_as(dw)
    ph = (y2 - y1).unsqueeze(-1).expand_as(dh)
    # Use exp(network energy) to enlarge/shrink each roi
    gw = pw * dw.exp()
    gh = ph * dh.exp()
    # Use network energy to shift the center of each roi
    gx = px + pw * dx
    gy = py + ph * dy
    # Convert center-xy/width/height to top-left, bottom-right
    x1 = gx - gw * 0.5
    y1 = gy - gh * 0.5
    x2 = gx + gw * 0.5
    y2 = gy + gh * 0.5

    bboxes = torch.stack([x1, y1, x2, y2], dim=-1).view(deltas.size())

    if clip_border and max_shape is not None:
        if not isinstance(max_shape, torch.Tensor):
            max_shape = x1.new_tensor(max_shape)
        max_shape = max_shape[..., :2].type_as(x1)
        if max_shape.ndim == 2:
            assert bboxes.ndim == 3
            assert max_shape.size(0) == bboxes.size(0)

        min_xy = x1.new_tensor(0)
        max_xy = torch.cat(
            [max_shape] * (deltas.size(-1) // 2),
            dim=-1).flip(-1).unsqueeze(-2)
        bboxes = torch.where(bboxes < min_xy, min_xy, bboxes)
        bboxes = torch.where(bboxes > max_xy, max_xy, bboxes)

    return bboxes



import mmcv
from mmcv.image import tensor2imgs

from mmdet.core import bbox_mapping
from ..builder import DETECTORS, build_backbone, build_head, build_neck
from .base import BaseDetector


@DETECTORS.register_module()
class RPN(BaseDetector):
    """Implementation of Region Proposal Network."""

    def __init__(self,
                 backbone,
                 neck,
                 rpn_head,
                 train_cfg,
                 test_cfg,
                 pretrained=None):
        super(RPN, self).__init__()
        self.backbone = build_backbone(backbone)
        self.neck = build_neck(neck) if neck is not None else None
        rpn_train_cfg = train_cfg.rpn if train_cfg is not None else None
        rpn_head.update(train_cfg=rpn_train_cfg)
        rpn_head.update(test_cfg=test_cfg.rpn)
        self.rpn_head = build_head(rpn_head)
        self.train_cfg = train_cfg
        self.test_cfg = test_cfg
        self.init_weights(pretrained=pretrained)

    def init_weights(self, pretrained=None):
        """Initialize the weights in detector.

        Args:
            pretrained (str, optional): Path to pre-trained weights.
                Defaults to None.
        """
        super(RPN, self).init_weights(pretrained)
        self.backbone.init_weights(pretrained=pretrained)
        if self.with_neck:
            self.neck.init_weights()
        self.rpn_head.init_weights()

    def extract_feat(self, img):
        """Extract features.

        Args:
            img (torch.Tensor): Image tensor with shape (n, c, h ,w).

        Returns:
            list[torch.Tensor]: Multi-level features that may have
                different resolutions.
        """
        x = self.backbone(img)
        if self.with_neck:
            x = self.neck(x)
        return x

    def forward_dummy(self, img):
        """Dummy forward function."""
        x = self.extract_feat(img)
        rpn_outs = self.rpn_head(x)
        return rpn_outs

    def forward_train(self,
                      img,
                      img_metas,
                      gt_bboxes=None,
                      gt_bboxes_ignore=None):
        """
        Args:
            img (Tensor): Input images of shape (N, C, H, W).
                Typically these should be mean centered and std scaled.
            img_metas (list[dict]): A List of image info dict where each dict
                has: 'img_shape', 'scale_factor', 'flip', and may also contain
                'filename', 'ori_shape', 'pad_shape', and 'img_norm_cfg'.
                For details on the values of these keys see
                :class:`mmdet.datasets.pipelines.Collect`.
            gt_bboxes (list[Tensor]): Each item are the truth boxes for each
                image in [tl_x, tl_y, br_x, br_y] format.
            gt_bboxes_ignore (None | list[Tensor]): Specify which bounding
                boxes can be ignored when computing the loss.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        if (isinstance(self.train_cfg.rpn, dict)
                and self.train_cfg.rpn.get('debug', False)):
            self.rpn_head.debug_imgs = tensor2imgs(img)

        x = self.extract_feat(img)
        losses = self.rpn_head.forward_train(x, img_metas, gt_bboxes, None,
                                             gt_bboxes_ignore)
        return losses

    def simple_test(self, img, img_metas, rescale=False):
        """Test function without test time augmentation.

        Args:
            imgs (list[torch.Tensor]): List of multiple images
            img_metas (list[dict]): List of image information.
            rescale (bool, optional): Whether to rescale the results.
                Defaults to False.

        Returns:
            list[np.ndarray]: proposals
        """
        x = self.extract_feat(img)
        proposal_list = self.rpn_head.simple_test_rpn(x, img_metas)
        if rescale:
            for proposals, meta in zip(proposal_list, img_metas):
                proposals[:, :4] /= proposals.new_tensor(meta['scale_factor'])

        return [proposal.cpu().numpy() for proposal in proposal_list]

    def aug_test(self, imgs, img_metas, rescale=False):
        """Test function with test time augmentation.

        Args:
            imgs (list[torch.Tensor]): List of multiple images
            img_metas (list[dict]): List of image information.
            rescale (bool, optional): Whether to rescale the results.
                Defaults to False.

        Returns:
            list[np.ndarray]: proposals
        """
        proposal_list = self.rpn_head.aug_test_rpn(
            self.extract_feats(imgs), img_metas)
        if not rescale:
            for proposals, img_meta in zip(proposal_list, img_metas[0]):
                img_shape = img_meta['img_shape']
                scale_factor = img_meta['scale_factor']
                flip = img_meta['flip']
                flip_direction = img_meta['flip_direction']
                proposals[:, :4] = bbox_mapping(proposals[:, :4], img_shape,
                                                scale_factor, flip,
                                                flip_direction)
        return [proposal.cpu().numpy() for proposal in proposal_list]

    def show_result(self, data, result, top_k=20, **kwargs):
        """Show RPN proposals on the image.

        Args:
            data (str or np.ndarray): Image filename or loaded image.
            result (Tensor or tuple): The results to draw over `img`
                bbox_result or (bbox_result, segm_result).
            top_k (int): Plot the first k bboxes only
               if set positive. Default: 20

        Returns:
            np.ndarray: The image with bboxes drawn on it.
        """
        mmcv.imshow_bboxes(data, result, top_k=top_k)



from torchvision.datasets.utils import download_url
from ldm.util import instantiate_from_config
import torch
import os
# todo ?
from google.colab import files
from IPython.display import Image as ipyimg
import ipywidgets as widgets
from PIL import Image
from numpy import asarray
from einops import rearrange, repeat
import torch, torchvision
from ldm.models.diffusion.ddim import DDIMSampler
from ldm.util import ismap
import time
from omegaconf import OmegaConf


def download_models(mode):

    if mode == "superresolution":
        # this is the small bsr light model
        url_conf = 'https://heibox.uni-heidelberg.de/f/31a76b13ea27482981b4/?dl=1'
        url_ckpt = 'https://heibox.uni-heidelberg.de/f/578df07c8fc04ffbadf3/?dl=1'

        path_conf = 'logs/diffusion/superresolution_bsr/configs/project.yaml'
        path_ckpt = 'logs/diffusion/superresolution_bsr/checkpoints/last.ckpt'

        download_url(url_conf, path_conf)
        download_url(url_ckpt, path_ckpt)

        path_conf = path_conf + '/?dl=1' # fix it
        path_ckpt = path_ckpt + '/?dl=1' # fix it
        return path_conf, path_ckpt

    else:
        raise NotImplementedError


def load_model_from_config(config, ckpt):
    print(f"Loading model from {ckpt}")
    pl_sd = torch.load(ckpt, map_location="cpu")
    global_step = pl_sd["global_step"]
    sd = pl_sd["state_dict"]
    model = instantiate_from_config(config.model)
    m, u = model.load_state_dict(sd, strict=False)
    model.cuda()
    model.eval()
    return {"model": model}, global_step


def get_model(mode):
    path_conf, path_ckpt = download_models(mode)
    config = OmegaConf.load(path_conf)
    model, step = load_model_from_config(config, path_ckpt)
    return model


def get_custom_cond(mode):
    dest = "data/example_conditioning"

    if mode == "superresolution":
        uploaded_img = files.upload()
        filename = next(iter(uploaded_img))
        name, filetype = filename.split(".") # todo assumes just one dot in name !
        os.rename(f"{filename}", f"{dest}/{mode}/custom_{name}.{filetype}")

    elif mode == "text_conditional":
        w = widgets.Text(value='A cake with cream!', disabled=True)
        display(w)

        with open(f"{dest}/{mode}/custom_{w.value[:20]}.txt", 'w') as f:
            f.write(w.value)

    elif mode == "class_conditional":
        w = widgets.IntSlider(min=0, max=1000)
        display(w)
        with open(f"{dest}/{mode}/custom.txt", 'w') as f:
            f.write(w.value)

    else:
        raise NotImplementedError(f"cond not implemented for mode{mode}")


def get_cond_options(mode):
    path = "data/example_conditioning"
    path = os.path.join(path, mode)
    onlyfiles = [f for f in sorted(os.listdir(path))]
    return path, onlyfiles


def select_cond_path(mode):
    path = "data/example_conditioning"  # todo
    path = os.path.join(path, mode)
    onlyfiles = [f for f in sorted(os.listdir(path))]

    selected = widgets.RadioButtons(
        options=onlyfiles,
        description='Select conditioning:',
        disabled=False
    )
    display(selected)
    selected_path = os.path.join(path, selected.value)
    return selected_path


def get_cond(mode, selected_path):
    example = dict()
    if mode == "superresolution":
        up_f = 4
        visualize_cond_img(selected_path)

        c = Image.open(selected_path)
        c = torch.unsqueeze(torchvision.transforms.ToTensor()(c), 0)
        c_up = torchvision.transforms.functional.resize(c, size=[up_f * c.shape[2], up_f * c.shape[3]], antialias=True)
        c_up = rearrange(c_up, '1 c h w -> 1 h w c')
        c = rearrange(c, '1 c h w -> 1 h w c')
        c = 2. * c - 1.

        c = c.to(torch.device("cuda"))
        example["LR_image"] = c
        example["image"] = c_up

    return example


def visualize_cond_img(path):
    display(ipyimg(filename=path))


def run(model, selected_path, task, custom_steps, resize_enabled=False, classifier_ckpt=None, global_step=None):

    example = get_cond(task, selected_path)

    save_intermediate_vid = False
    n_runs = 1
    masked = False
    guider = None
    ckwargs = None
    mode = 'ddim'
    ddim_use_x0_pred = False
    temperature = 1.
    eta = 1.
    make_progrow = True
    custom_shape = None

    height, width = example["image"].shape[1:3]
    split_input = height >= 128 and width >= 128

    if split_input:
        ks = 128
        stride = 64
        vqf = 4  #
        model.split_input_params = {"ks": (ks, ks), "stride": (stride, stride),
                                    "vqf": vqf,
                                    "patch_distributed_vq": True,
                                    "tie_braker": False,
                                    "clip_max_weight": 0.5,
                                    "clip_min_weight": 0.01,
                                    "clip_max_tie_weight": 0.5,
                                    "clip_min_tie_weight": 0.01}
    else:
        if hasattr(model, "split_input_params"):
            delattr(model, "split_input_params")

    invert_mask = False

    x_T = None
    for n in range(n_runs):
        if custom_shape is not None:
            x_T = torch.randn(1, custom_shape[1], custom_shape[2], custom_shape[3]).to(model.device)
            x_T = repeat(x_T, '1 c h w -> b c h w', b=custom_shape[0])

        logs = make_convolutional_sample(example, model,
                                         mode=mode, custom_steps=custom_steps,
                                         eta=eta, swap_mode=False , masked=masked,
                                         invert_mask=invert_mask, quantize_x0=False,
                                         custom_schedule=None, decode_interval=10,
                                         resize_enabled=resize_enabled, custom_shape=custom_shape,
                                         temperature=temperature, noise_dropout=0.,
                                         corrector=guider, corrector_kwargs=ckwargs, x_T=x_T, save_intermediate_vid=save_intermediate_vid,
                                         make_progrow=make_progrow,ddim_use_x0_pred=ddim_use_x0_pred
                                         )
    return logs


@torch.no_grad()
def convsample_ddim(model, cond, steps, shape, eta=1.0, callback=None, normals_sequence=None,
                    mask=None, x0=None, quantize_x0=False, img_callback=None,
                    temperature=1., noise_dropout=0., score_corrector=None,
                    corrector_kwargs=None, x_T=None, log_every_t=None
                    ):

    ddim = DDIMSampler(model)
    bs = shape[0]  # dont know where this comes from but wayne
    shape = shape[1:]  # cut batch dim
    print(f"Sampling with eta = {eta}; steps: {steps}")
    samples, intermediates = ddim.sample(steps, batch_size=bs, shape=shape, conditioning=cond, callback=callback,
                                         normals_sequence=normals_sequence, quantize_x0=quantize_x0, eta=eta,
                                         mask=mask, x0=x0, temperature=temperature, verbose=False,
                                         score_corrector=score_corrector,
                                         corrector_kwargs=corrector_kwargs, x_T=x_T)

    return samples, intermediates


@torch.no_grad()
def make_convolutional_sample(batch, model, mode="vanilla", custom_steps=None, eta=1.0, swap_mode=False, masked=False,
                              invert_mask=True, quantize_x0=False, custom_schedule=None, decode_interval=1000,
                              resize_enabled=False, custom_shape=None, temperature=1., noise_dropout=0., corrector=None,
                              corrector_kwargs=None, x_T=None, save_intermediate_vid=False, make_progrow=True,ddim_use_x0_pred=False):
    log = dict()

    z, c, x, xrec, xc = model.get_input(batch, model.first_stage_key,
                                        return_first_stage_outputs=True,
                                        force_c_encode=not (hasattr(model, 'split_input_params')
                                                            and model.cond_stage_key == 'coordinates_bbox'),
                                        return_original_cond=True)

    log_every_t = 1 if save_intermediate_vid else None

    if custom_shape is not None:
        z = torch.randn(custom_shape)
        print(f"Generating {custom_shape[0]} samples of shape {custom_shape[1:]}")

    z0 = None

    log["input"] = x
    log["reconstruction"] = xrec

    if ismap(xc):
        log["original_conditioning"] = model.to_rgb(xc)
        if hasattr(model, 'cond_stage_key'):
            log[model.cond_stage_key] = model.to_rgb(xc)

    else:
        log["original_conditioning"] = xc if xc is not None else torch.zeros_like(x)
        if model.cond_stage_model:
            log[model.cond_stage_key] = xc if xc is not None else torch.zeros_like(x)
            if model.cond_stage_key =='class_label':
                log[model.cond_stage_key] = xc[model.cond_stage_key]

    with model.ema_scope("Plotting"):
        t0 = time.time()
        img_cb = None

        sample, intermediates = convsample_ddim(model, c, steps=custom_steps, shape=z.shape,
                                                eta=eta,
                                                quantize_x0=quantize_x0, img_callback=img_cb, mask=None, x0=z0,
                                                temperature=temperature, noise_dropout=noise_dropout,
                                                score_corrector=corrector, corrector_kwargs=corrector_kwargs,
                                                x_T=x_T, log_every_t=log_every_t)
        t1 = time.time()

        if ddim_use_x0_pred:
            sample = intermediates['pred_x0'][-1]

    x_sample = model.decode_first_stage(sample)

    try:
        x_sample_noquant = model.decode_first_stage(sample, force_not_quantize=True)
        log["sample_noquant"] = x_sample_noquant
        log["sample_diff"] = torch.abs(x_sample_noquant - x_sample)
    except:
        pass

    log["sample"] = x_sample
    log["time"] = t1 - t0

    return log



import copy
import warnings

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcv import ConfigDict
from mmcv.cnn import normal_init
from mmcv.ops import batched_nms

from ..builder import HEADS
from .anchor_head import AnchorHead
from .rpn_test_mixin import RPNTestMixin


@HEADS.register_module()
class RPNHead(RPNTestMixin, AnchorHead):
    """RPN head.

    Args:
        in_channels (int): Number of channels in the input feature map.
    """  # noqa: W605

    def __init__(self, in_channels, **kwargs):
        super(RPNHead, self).__init__(1, in_channels, **kwargs)

    def _init_layers(self):
        """Initialize layers of the head."""
        self.rpn_conv = nn.Conv2d(
            self.in_channels, self.feat_channels, 3, padding=1)
        self.rpn_cls = nn.Conv2d(self.feat_channels,
                                 self.num_anchors * self.cls_out_channels, 1)
        self.rpn_reg = nn.Conv2d(self.feat_channels, self.num_anchors * 4, 1)

    def init_weights(self):
        """Initialize weights of the head."""
        normal_init(self.rpn_conv, std=0.01)
        normal_init(self.rpn_cls, std=0.01)
        normal_init(self.rpn_reg, std=0.01)

    def forward_single(self, x):
        """Forward feature map of a single scale level."""
        x = self.rpn_conv(x)
        x = F.relu(x, inplace=True)
        rpn_cls_score = self.rpn_cls(x)
        rpn_bbox_pred = self.rpn_reg(x)
        return rpn_cls_score, rpn_bbox_pred

    def loss(self,
             cls_scores,
             bbox_preds,
             gt_bboxes,
             img_metas,
             gt_bboxes_ignore=None):
        """Compute losses of the head.

        Args:
            cls_scores (list[Tensor]): Box scores for each scale level
                Has shape (N, num_anchors * num_classes, H, W)
            bbox_preds (list[Tensor]): Box energies / deltas for each scale
                level with shape (N, num_anchors * 4, H, W)
            gt_bboxes (list[Tensor]): Ground truth bboxes for each image with
                shape (num_gts, 4) in [tl_x, tl_y, br_x, br_y] format.
            img_metas (list[dict]): Meta information of each image, e.g.,
                image size, scaling factor, etc.
            gt_bboxes_ignore (None | list[Tensor]): specify which bounding
                boxes can be ignored when computing the loss.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        losses = super(RPNHead, self).loss(
            cls_scores,
            bbox_preds,
            gt_bboxes,
            None,
            img_metas,
            gt_bboxes_ignore=gt_bboxes_ignore)
        return dict(
            loss_rpn_cls=losses['loss_cls'], loss_rpn_bbox=losses['loss_bbox'])

    def _get_bboxes(self,
                    cls_scores,
                    bbox_preds,
                    mlvl_anchors,
                    img_shapes,
                    scale_factors,
                    cfg,
                    rescale=False):
        """Transform outputs for a single batch item into bbox predictions.

        Args:
            cls_scores (list[Tensor]): Box scores for each scale level
                Has shape (N, num_anchors * num_classes, H, W).
            bbox_preds (list[Tensor]): Box energies / deltas for each scale
                level with shape (N, num_anchors * 4, H, W).
            mlvl_anchors (list[Tensor]): Box reference for each scale level
                with shape (num_total_anchors, 4).
            img_shapes (list[tuple[int]]): Shape of the input image,
                (height, width, 3).
            scale_factors (list[ndarray]): Scale factor of the image arange as
                (w_scale, h_scale, w_scale, h_scale).
            cfg (mmcv.Config): Test / postprocessing configuration,
                if None, test_cfg would be used.
            rescale (bool): If True, return boxes in original image space.

        Returns:
            list[tuple[Tensor, Tensor]]: Each item in result_list is 2-tuple.
                The first item is an (n, 5) tensor, where the first 4 columns
                are bounding box positions (tl_x, tl_y, br_x, br_y) and the
                5-th column is a score between 0 and 1. The second item is a
                (n,) tensor where each item is the predicted class labelof the
                corresponding box.
        """
        cfg = self.test_cfg if cfg is None else cfg
        cfg = copy.deepcopy(cfg)
        # bboxes from different level should be independent during NMS,
        # level_ids are used as labels for batched NMS to separate them
        level_ids = []
        mlvl_scores = []
        mlvl_bbox_preds = []
        mlvl_valid_anchors = []
        batch_size = cls_scores[0].shape[0]
        nms_pre_tensor = torch.tensor(
            cfg.nms_pre, device=cls_scores[0].device, dtype=torch.long)
        for idx in range(len(cls_scores)):
            rpn_cls_score = cls_scores[idx]
            rpn_bbox_pred = bbox_preds[idx]
            assert rpn_cls_score.size()[-2:] == rpn_bbox_pred.size()[-2:]
            rpn_cls_score = rpn_cls_score.permute(0, 2, 3, 1)
            if self.use_sigmoid_cls:
                rpn_cls_score = rpn_cls_score.reshape(batch_size, -1)
                scores = rpn_cls_score.sigmoid()
            else:
                rpn_cls_score = rpn_cls_score.reshape(batch_size, -1, 2)
                # We set FG labels to [0, num_class-1] and BG label to
                # num_class in RPN head since mmdet v2.5, which is unified to
                # be consistent with other head since mmdet v2.0. In mmdet v2.0
                # to v2.4 we keep BG label as 0 and FG label as 1 in rpn head.
                scores = rpn_cls_score.softmax(-1)[..., 0]
            rpn_bbox_pred = rpn_bbox_pred.permute(0, 2, 3, 1).reshape(
                batch_size, -1, 4)
            anchors = mlvl_anchors[idx]
            anchors = anchors.expand_as(rpn_bbox_pred)
            if nms_pre_tensor > 0:
                # sort is faster than topk
                # _, topk_inds = scores.topk(cfg.nms_pre)
                # keep topk op for dynamic k in onnx model
                if torch.onnx.is_in_onnx_export():
                    # sort op will be converted to TopK in onnx
                    # and k<=3480 in TensorRT
                    scores_shape = torch._shape_as_tensor(scores)
                    nms_pre = torch.where(scores_shape[1] < nms_pre_tensor,
                                          scores_shape[1], nms_pre_tensor)
                    _, topk_inds = scores.topk(nms_pre)
                    batch_inds = torch.arange(batch_size).view(
                        -1, 1).expand_as(topk_inds)
                    scores = scores[batch_inds, topk_inds]
                    rpn_bbox_pred = rpn_bbox_pred[batch_inds, topk_inds, :]
                    anchors = anchors[batch_inds, topk_inds, :]

                elif scores.shape[-1] > cfg.nms_pre:
                    ranked_scores, rank_inds = scores.sort(descending=True)
                    topk_inds = rank_inds[:, :cfg.nms_pre]
                    scores = ranked_scores[:, :cfg.nms_pre]
                    batch_inds = torch.arange(batch_size).view(
                        -1, 1).expand_as(topk_inds)
                    rpn_bbox_pred = rpn_bbox_pred[batch_inds, topk_inds, :]
                    anchors = anchors[batch_inds, topk_inds, :]

            mlvl_scores.append(scores)
            mlvl_bbox_preds.append(rpn_bbox_pred)
            mlvl_valid_anchors.append(anchors)
            level_ids.append(
                scores.new_full((
                    batch_size,
                    scores.size(1),
                ),
                                idx,
                                dtype=torch.long))

        batch_mlvl_scores = torch.cat(mlvl_scores, dim=1)
        batch_mlvl_anchors = torch.cat(mlvl_valid_anchors, dim=1)
        batch_mlvl_rpn_bbox_pred = torch.cat(mlvl_bbox_preds, dim=1)
        batch_mlvl_proposals = self.bbox_coder.decode(
            batch_mlvl_anchors, batch_mlvl_rpn_bbox_pred, max_shape=img_shapes)
        batch_mlvl_ids = torch.cat(level_ids, dim=1)

        # deprecate arguments warning
        if 'nms' not in cfg or 'max_num' in cfg or 'nms_thr' in cfg:
            warnings.warn(
                'In rpn_proposal or test_cfg, '
                'nms_thr has been moved to a dict named nms as '
                'iou_threshold, max_num has been renamed as max_per_img, '
                'name of original arguments and the way to specify '
                'iou_threshold of NMS will be deprecated.')
        if 'nms' not in cfg:
            cfg.nms = ConfigDict(dict(type='nms', iou_threshold=cfg.nms_thr))
        if 'max_num' in cfg:
            if 'max_per_img' in cfg:
                assert cfg.max_num == cfg.max_per_img, f'You ' \
                    f'set max_num and ' \
                    f'max_per_img at the same time, but get {cfg.max_num} ' \
                    f'and {cfg.max_per_img} respectively' \
                    'Please delete max_num which will be deprecated.'
            else:
                cfg.max_per_img = cfg.max_num
        if 'nms_thr' in cfg:
            assert cfg.nms.iou_threshold == cfg.nms_thr, f'You set' \
                f' iou_threshold in nms and ' \
                f'nms_thr at the same time, but get' \
                f' {cfg.nms.iou_threshold} and {cfg.nms_thr}' \
                f' respectively. Please delete the nms_thr ' \
                f'which will be deprecated.'

        result_list = []
        for (mlvl_proposals, mlvl_scores,
             mlvl_ids) in zip(batch_mlvl_proposals, batch_mlvl_scores,
                              batch_mlvl_ids):
            # Skip nonzero op while exporting to ONNX
            if cfg.min_bbox_size > 0 and (not torch.onnx.is_in_onnx_export()):
                w = mlvl_proposals[:, 2] - mlvl_proposals[:, 0]
                h = mlvl_proposals[:, 3] - mlvl_proposals[:, 1]
                valid_ind = torch.nonzero(
                    (w >= cfg.min_bbox_size)
                    & (h >= cfg.min_bbox_size),
                    as_tuple=False).squeeze()
                if valid_ind.sum().item() != len(mlvl_proposals):
                    mlvl_proposals = mlvl_proposals[valid_ind, :]
                    mlvl_scores = mlvl_scores[valid_ind]
                    mlvl_ids = mlvl_ids[valid_ind]

            dets, keep = batched_nms(mlvl_proposals, mlvl_scores, mlvl_ids,
                                     cfg.nms)
            result_list.append(dets[:cfg.max_per_img])
        return result_list



# Copyright (c) OpenMMLab. All rights reserved.
from itertools import chain

from torch.nn.parallel import DataParallel

from .scatter_gather import scatter_kwargs


class MMDataParallel(DataParallel):
    """The DataParallel module that supports DataContainer.

    MMDataParallel has two main differences with PyTorch DataParallel:

    - It supports a custom type :class:`DataContainer` which allows more
      flexible control of input data during both GPU and CPU inference.
    - It implement two more APIs ``train_step()`` and ``val_step()``.

    Args:
        module (:class:`nn.Module`): Module to be encapsulated.
        device_ids (list[int]): Device IDS of modules to be scattered to.
            Defaults to None when GPU is not available.
        output_device (str | int): Device ID for output. Defaults to None.
        dim (int): Dimension used to scatter the data. Defaults to 0.
    """

    def __init__(self, *args, dim=0, **kwargs):
        super(MMDataParallel, self).__init__(*args, dim=dim, **kwargs)
        self.dim = dim

    def forward(self, *inputs, **kwargs):
        """Override the original forward function.

        The main difference lies in the CPU inference where the data in
        :class:`DataContainers` will still be gathered.
        """
        if not self.device_ids:
            # We add the following line thus the module could gather and
            # convert data containers as those in GPU inference
            inputs, kwargs = self.scatter(inputs, kwargs, [-1])
            return self.module(*inputs[0], **kwargs[0])
        else:
            return super().forward(*inputs, **kwargs)

    def scatter(self, inputs, kwargs, device_ids):
        return scatter_kwargs(inputs, kwargs, device_ids, dim=self.dim)

    def train_step(self, *inputs, **kwargs):
        if not self.device_ids:
            # We add the following line thus the module could gather and
            # convert data containers as those in GPU inference
            inputs, kwargs = self.scatter(inputs, kwargs, [-1])
            return self.module.train_step(*inputs[0], **kwargs[0])

        assert len(self.device_ids) == 1, \
            ('MMDataParallel only supports single GPU training, if you need to'
             ' train with multiple GPUs, please use MMDistributedDataParallel'
             'instead.')

        for t in chain(self.module.parameters(), self.module.buffers()):
            if t.device != self.src_device_obj:
                raise RuntimeError(
                    'module must have its parameters and buffers '
                    f'on device {self.src_device_obj} (device_ids[0]) but '
                    f'found one of them on device: {t.device}')

        inputs, kwargs = self.scatter(inputs, kwargs, self.device_ids)
        return self.module.train_step(*inputs[0], **kwargs[0])

    def val_step(self, *inputs, **kwargs):
        if not self.device_ids:
            # We add the following line thus the module could gather and
            # convert data containers as those in GPU inference
            inputs, kwargs = self.scatter(inputs, kwargs, [-1])
            return self.module.val_step(*inputs[0], **kwargs[0])

        assert len(self.device_ids) == 1, \
            ('MMDataParallel only supports single GPU training, if you need to'
             ' train with multiple GPUs, please use MMDistributedDataParallel'
             ' instead.')

        for t in chain(self.module.parameters(), self.module.buffers()):
            if t.device != self.src_device_obj:
                raise RuntimeError(
                    'module must have its parameters and buffers '
                    f'on device {self.src_device_obj} (device_ids[0]) but '
                    f'found one of them on device: {t.device}')

        inputs, kwargs = self.scatter(inputs, kwargs, self.device_ids)
        return self.module.val_step(*inputs[0], **kwargs[0])



# modified from https://github.com/DreamEditBenchTeam/DreamEdit/blob/main/src/iterate_generate.py
import warnings

warnings.filterwarnings("ignore")

import os
from omegaconf import OmegaConf
import argparse
import torchvision
import torchvision.transforms as T
import copy
import glob
from tqdm import tqdm
import numpy as np
from PIL import Image
import torch

from imagen_hub.miscmodels.blip import BLIP_Model
from imagen_hub.miscmodels import LangSAM, CLIP, VITs16
from imagen_hub.depend.lang_sam.lang_sam import draw_image
from imagen_hub.metrics.dreambooth_metric import evaluate_dino_score, evaluate_clipi_score, evaluate_clipi_score_list,evaluate_dino_score_list
from imagen_hub.utils.save_image_helper import get_concat_pil_images, get_mask_pil_image, save_pil_image

from .generate import load_model_from_config, diffedit

from .utils.mask_helper import merge_masks, subtract_mask, get_polished_mask, mask_dilation, mask_closing, mask_closing_half, transform_box_mask, transform_box_mask_paste, resize_box_from_middle, resize_box_from_bottom, bounding_box_merge, get_mask_from_bbox
from .pipelines.inpainting_pipelines import select_inpainting_pipeline, inpaint_text_gligen
from .pipelines.extract_object_pipeline import get_object_caption

import json

import logging

logging.getLogger().setLevel(logging.ERROR)

class EncodeRatioScheduler:
    def __init__(self, schedule_config, total_iterations):
        self.start_ratio = schedule_config.start_ratio
        self.end_ratio = schedule_config.end_ratio
        self.decay_type = schedule_config.decay_type
        self.current_ratio = self.start_ratio
        self.manual_ratio_list = []
        if self.decay_type == 'linear':
            self.decay_rate = (self.start_ratio - self.end_ratio) / total_iterations
        elif self.decay_type == 'exponential':
            self.decay_rate = np.power(self.end_ratio / self.start_ratio, 1 / total_iterations)
        elif self.decay_type == 'manual':
            self.decay_rate = 0
            self.manual_ratio_list = copy.deepcopy(schedule_config.manual_ratio_list)
        else:  # constant
            self.decay_rate = 0

    def step(self):
        if self.decay_type == 'linear':
            self.current_ratio -= self.decay_rate
        elif self.decay_type == 'exponential':
            self.current_ratio *= self.decay_rate
        elif self.decay_type == 'manual':
            self.current_ratio = self.manual_ratio_list.pop(0)
        else:  # constant
            self.current_ratio = self.start_ratio
        print("Current encode ratio: ", self.current_ratio)
        return self.current_ratio


def replace_background(
        src_img,
        img_to_edit,
        obj_class,
        first_iter_obj_mask,
        lang_sam_model,
        lang_sam_config,
        de_config,
        difference = False
):
    assert src_img.size == img_to_edit.size
    bg_config = de_config.background_correction
    if bg_config.use_obj_mask_from_first_iteration and first_iter_obj_mask is not None:
        obj_mask = first_iter_obj_mask
        if difference:
            masks, boxes, phrases, logits = lang_sam_model.predict(
                img_to_edit,
                obj_class,
                box_threshold=lang_sam_config.segment_confidence,
                text_threshold=lang_sam_config.segment_confidence,
            )
            obj_mask = merge_masks(masks).long()
            obj_mask = subtract_mask(first_iter_obj_mask, obj_mask).bool() == False
            obj_mask = obj_mask.long()
    else:
        masks, boxes, phrases, logits = lang_sam_model.predict(
            img_to_edit,
            obj_class,
            box_threshold=lang_sam_config.segment_confidence,
            text_threshold=lang_sam_config.segment_confidence,
        )
        obj_mask = merge_masks(masks)
        obj_mask = get_polished_mask(obj_mask, 2, "dilation")
    # mask_config = de_config.mask_config
    # obj_mask = get_polished_mask(obj_mask, mask_config.mask_dilate_kernel, mask_config.mask_type)
    obj_mask = obj_mask.unsqueeze(-1).repeat(1, 1, 3)
    combined = torch.where(
        obj_mask > 0,
        torch.from_numpy(np.asarray(img_to_edit)),
        torch.from_numpy(np.asarray(src_img))
    ).detach().cpu().numpy()
    img_edit = Image.fromarray(np.uint8(combined))
    return img_edit


def iterative_edit(
        src_img_path,  # background image
        img_to_edit_path,
        obj_class,
        lang_sam_model,
        sd_model,
        src_prompt,
        dst_prompt,
        config,
        device,
        labeled_box=None,
):
    src_img = Image.open(src_img_path).resize((512, 512))
    src_img_edited = Image.open(img_to_edit_path).resize((512, 512))
    de_config = config.model.de
    lang_sam_config = config.model.lang_sam
    segment_confidence = lang_sam_config.segment_confidence
    ddim_config = de_config.ddim

    # Step1. Iteratively dream edit
    intermediate_images = []
    obj_mask_list = []
    latents_record_list = []
    generated_image_list = []
    first_iter_obj_mask = None
    encode_ratio_scheduler = EncodeRatioScheduler(ddim_config.encode_ratio_schedule, ddim_config.iteration_number)

    if de_config.use_diffedit:
        pbar = tqdm(range(1))
    else:
        pbar = tqdm(range(ddim_config.iteration_number))
    for i in pbar:
        if de_config.use_diffedit:
            pbar.set_description("Diffedit iteration %s" % i)
        else:
            pbar.set_description("Dream edit iteration %s" % i)

        img_to_edit = Image.open(img_to_edit_path)  # type, PIL image
        if img_to_edit.size != (512, 512):
            img_to_edit = img_to_edit.resize((512, 512))

        # Get object mask
        mask_config = de_config.mask_config
        if (mask_config.use_bbox_mask_for_first_iteration and i == 0) or mask_config.use_bbox_mask_for_all_iterations:
            obj_mask = get_mask_from_bbox(labeled_box).long()

            obj_masks, obj_boxes, phrases, logits = lang_sam_model.predict(
                img_to_edit,
                obj_class,
                box_threshold=segment_confidence,
                text_threshold=segment_confidence,
            )
            obj_mask_org = merge_masks(obj_masks)
            bound_mask = get_mask_from_bbox(labeled_box)
            labeled_img_to_edit = draw_image(
                np.asarray(img_to_edit),
                torch.unsqueeze(bound_mask, 0),
                torch.Tensor([labeled_box]),
                ['label'],
            )
        else:
            # Get the mask of the object from the previous iteration image
            obj_masks, obj_boxes, phrases, logits = lang_sam_model.predict(
                img_to_edit,
                obj_class,
                box_threshold=segment_confidence,
                text_threshold=segment_confidence,
            )
            obj_mask = merge_masks(obj_masks)
            obj_mask_org = obj_mask
            # obj_mask_org = get_polished_mask(obj_mask_org, 2, mask_config.mask_type)
            # obj_mask = get_polished_mask(obj_mask, mask_config.mask_dilate_kernel, mask_config.mask_type)
            obj_mask = get_polished_mask(obj_mask, mask_config.mask_dilate_kernel, mask_config.mask_type)
            labels = [f"{phrase} {logit:.2f}" for phrase, logit in zip(phrases, logits)]
            labeled_img_to_edit = draw_image(
                np.asarray(img_to_edit),
                obj_masks,
                obj_boxes,
                labels
            )

        # Save intermediate img_to_edit and unpolished obj_mask
        labeled_img_to_edit = Image.fromarray(np.uint8(labeled_img_to_edit)).convert("RGB")
        intermediate_images.append(labeled_img_to_edit)
        obj_mask_list.append(obj_mask_org)

        first_iter_obj_mask = obj_mask if i == 0 else first_iter_obj_mask

        # Dream edit
        encode_ratio = encode_ratio_scheduler.step()
        transform = T.Resize((512 // 8, 512 // 8))
        latent_obj_mask = transform(obj_mask[None, ...])[0].to(device)
        _img_edited, img_to_edit_reconstruct, latents_record_list, diffedit_mask = diffedit(
            model=sd_model,
            config=de_config,
            init_image=img_to_edit_path,
            iteration=i,
            src_prompt=src_prompt,
            dst_prompt=dst_prompt,
            encode_ratio=encode_ratio,
            ddim_steps=ddim_config.ddim_steps,
            scale=ddim_config.scale,
            seed=ddim_config.seed,
            sam_mask=latent_obj_mask,
            H=512,
            W=512,
            noise_step=ddim_config.noise_step,
            record_list=latents_record_list,
            use_diffedit=de_config.use_diffedit
        )
        img_edited = Image.fromarray(_img_edited[0])  # type, PIL image

        # Replace the background with the original one
        # previous_masks_merge = merge_masks(torch.stack(obj_mask_list, dim=0))
        # process_mask_difference = subtract_mask(previous_masks_merge, obj_mask_list[-1])
        # if de_config.background_correction_enabled and i <= de_config.background_correction.iteration_number:
        if de_config.background_correction_enabled and i <= de_config.background_correction.iteration_number:
            if not de_config.use_diffedit:
                img_edited = replace_background(
                    src_img,
                    img_edited,
                    obj_class,
                    first_iter_obj_mask,
                    lang_sam_model,
                    lang_sam_config,
                    de_config,
                )

        # Save intermediate mask
        obj_mask_pil = get_mask_pil_image(obj_mask).convert('RGB')  # TODO: why size is not 512?
        obj_mask_pil = obj_mask_pil.resize(img_to_edit.size)
        intermediate_images.append(obj_mask_pil)

        if de_config.use_diffedit:
            diffedit_mask_pil = get_mask_pil_image(diffedit_mask.detach().cpu()).convert('RGB')  # TODO: why size is not 512?
            diffedit_mask_pil = diffedit_mask_pil.resize(img_to_edit.size)
            intermediate_images.append(diffedit_mask_pil)

        # Save intermediate edited image
        tmp_img_name = "tmp_" + config.data.db_folder_name + "_" + config.model.de.task_type
        save_pil_image(pil_image=img_edited, dest_folder="tmp", filename=tmp_img_name, filetype="jpg")
        generated_image_list.append(img_edited)
        img_to_edit_path = os.path.join("tmp", tmp_img_name + ".jpg")
        intermediate_images.append(img_edited)

    # Step2. Postprocessing
    if de_config.inpaint_after_last_iteration:
        # Remove last mask, then postprocessing
        curr_mask = obj_mask_list.pop()
        previous_masks_merge = merge_masks(torch.stack(obj_mask_list, dim=0))
        process_mask = subtract_mask(previous_masks_merge, curr_mask)
        process_mask_pil = get_mask_pil_image(process_mask).convert('RGB')
        intermediate_images.append(process_mask_pil)
        # Inpaint
        inpaint_pipe = select_inpainting_pipeline("sd_inpaint", device)
        img_edited = inpaint_pipe(
            prompt="",
            image=img_edited,
            mask_image=process_mask_pil
        ).images[0]
        intermediate_images.append(img_edited)

    return img_edited, intermediate_images, generated_image_list


def addition(
        src_img_path,
        obj_img_path,  # Useful only for copy-paste
        obj_class,
        lang_sam_model,
        sd_model,
        src_prompt,
        dst_prompt,
        config,
        device,
        labeled_box=None,
):
    de_config = config.model.de
    lang_sam_config = config.model.lang_sam
    segment_confidence = lang_sam_config.segment_confidence

    # Define tmp image path
    tmp_img_name = "tmp_" + config.data.db_folder_name + "_addition"
    tmp_img_path = os.path.join("tmp", tmp_img_name + ".jpg")
    if not os.path.exists(os.path.join("tmp", tmp_img_name)):
        os.makedirs(os.path.join("tmp", tmp_img_name))

    # Step1. Inpaint a sample object into the image
    src_img = Image.open(src_img_path).resize((512, 512))
    intermediate_images = [src_img]
    addition_config = de_config.addition_config

    if addition_config.automate_prompt:
        blip = BLIP_Model()
        caption = get_object_caption(src_img, obj_class, blip, lang_sam_model)
        addition_config_inpaint_prompt = caption
        addition_config_inpaint_phrase = caption
    else:
        addition_config_inpaint_prompt = addition_config.inpaint_prompt
        addition_config_inpaint_phrase = addition_config.inpaint_phrase

    if addition_config.use_copy_paste:
        obj_img = Image.open(obj_img_path).resize((512, 512))
        obj_masks, obj_boxes, phrases, logits = lang_sam_model.predict(
            obj_img,
            obj_class,
            box_threshold=segment_confidence,
            text_threshold=segment_confidence,
        )
        obj_mask = merge_masks(obj_masks)
        mask_config = de_config.mask_config
        # obj_mask = get_polished_mask(obj_mask, mask_config.mask_dilate_kernel, mask_config.mask_type)
        obj_box = bounding_box_merge(obj_boxes)
        obj_mask, inpainted_img = transform_box_mask_paste(labeled_box, obj_box, obj_mask, src_img, obj_img, )
        inpainted_img = Image.fromarray(np.uint8(inpainted_img))
        save_pil_image(pil_image=inpainted_img, dest_folder="tmp", filename=tmp_img_name, filetype="jpg")
        intermediate_images.append(inpainted_img)

    elif addition_config.inpaint_type == "stable_diffusion":
        obj_mask = get_mask_from_bbox(labeled_box).long()
        inpaint_pipe = select_inpainting_pipeline("sd_inpaint", device)
        inpainted_img = inpaint_pipe(
            prompt=addition_config_inpaint_prompt,
            image=src_img_path,
            mask_image=get_mask_pil_image(obj_mask).convert('RGB')
        ).images[0]
        save_pil_image(pil_image=inpainted_img, dest_folder="tmp", filename=tmp_img_name, filetype="jpg")
        intermediate_images.append(inpainted_img)

    elif addition_config.inpaint_type == "gligen":
        # TODO: look into this
        inpaint_pipe = select_inpainting_pipeline("gligen_inpaint", device)
        inpainted_imgs = inpaint_text_gligen(
            pipe=inpaint_pipe,
            prompt=addition_config_inpaint_prompt,
            background_path=src_img_path,
            bounding_box=labeled_box,
            gligen_phrase=addition_config_inpaint_phrase,
            config=config.model.gligen,
        )
        inpainted_imgs = torch.stack(
            [torch.from_numpy(image) for image in inpainted_imgs]
        ).permute(0, 3, 1, 2)
        torchvision.utils.save_image(inpainted_imgs, tmp_img_path, nrow=2, normalize=False)

        # Remove distorted background
        inpainted_img = Image.open(tmp_img_path).resize((512, 512))
        inpainted_img_bg_replaced = inpainted_img
        # I comment the intermediate rectification after gligen as it brings sharp edge
        # inpainted_img_bg_replaced = replace_background(
        #     src_img,
        #     inpainted_img,
        #     obj_class,
        #     None,
        #     lang_sam_model,
        #     lang_sam_config,
        #     de_config,
        # )
        save_pil_image(pil_image=inpainted_img_bg_replaced, dest_folder="tmp", filename=tmp_img_name, filetype="jpg")
        intermediate_images.append(inpainted_img)
        intermediate_images.append(inpainted_img_bg_replaced)
    else:
        raise NotImplementedError

    # Step2. Dream edit the inpainted image
    img_edited, intermediate_images_edited, generated_image_list = iterative_edit(
        src_img_path=src_img_path,
        img_to_edit_path=tmp_img_path,
        obj_class=obj_class,
        lang_sam_model=lang_sam_model,
        sd_model=sd_model,
        src_prompt=src_prompt,
        dst_prompt=dst_prompt,
        config=config,
        device=device,
        labeled_box=labeled_box,
    )
    intermediate_images = intermediate_images + intermediate_images_edited
    intermediate_images = get_concat_pil_images(intermediate_images, direction="h")
    return img_edited, intermediate_images, generated_image_list


def replacement(
        src_img_path,
        obj_img_path,  # Useful only for copy-paste
        obj_class,
        lang_sam_model,
        sd_model,
        src_prompt,
        dst_prompt,
        config,
        device,
):
    # TODO: implement background replacement
    src_img = Image.open(src_img_path).resize((512, 512))
    intermediate_images = [src_img]
    de_config = config.model.de
    lang_sam_config = config.model.lang_sam
    segment_confidence = lang_sam_config.segment_confidence
    replacement_config = de_config.addition_config
    tmp_img_name = "tmp_" + config.data.db_folder_name + "_replacement"
    tmp_img_path = os.path.join("tmp", tmp_img_name + ".jpg")

    if replacement_config.use_copy_paste:
        obj_img = Image.open(obj_img_path).resize((512, 512))
        obj_masks, obj_boxes, phrases, logits = lang_sam_model.predict(
            obj_img,
            obj_class,
            box_threshold=segment_confidence,
            text_threshold=segment_confidence,
        )
        obj_mask = merge_masks(obj_masks)
        obj_box = bounding_box_merge(obj_boxes)

        src_masks, src_boxes, phrases, logits = lang_sam_model.predict(
            src_img,
            obj_class,
            box_threshold=segment_confidence,
            text_threshold=segment_confidence,
        )
        src_mask = merge_masks(src_masks)
        src_box = bounding_box_merge(src_boxes)
        obj_mask, inpainted_img = transform_box_mask_paste(src_box, obj_box, obj_mask, src_img, obj_img, )
        inpainted_img = Image.fromarray(np.uint8(inpainted_img))
        save_pil_image(pil_image=inpainted_img, dest_folder="tmp", filename=tmp_img_name, filetype="jpg")
        intermediate_images.append(inpainted_img)
    else:
        save_pil_image(pil_image=src_img, dest_folder="tmp", filename=tmp_img_name, filetype="jpg")

    img_edited, intermediate_images_edited, generated_image_list = iterative_edit(
        src_img_path=src_img_path,
        img_to_edit_path=tmp_img_path,
        obj_class=obj_class,
        lang_sam_model=lang_sam_model,
        sd_model=sd_model,
        src_prompt=src_prompt,
        dst_prompt=dst_prompt,
        config=config,
        device=device,
    )
    intermediate_images = intermediate_images + intermediate_images_edited
    intermediate_images = get_concat_pil_images(intermediate_images, direction="h")
    return img_edited, intermediate_images, generated_image_list


def dream_edit(
        src_img_path,
        device,
        config,
        obj_img_path=None,
        bbox_file_path=None,
        sd_model=None,
        lang_sam_model=None,
        dino_model=None,
        clip_model=None
):
    src_img_file_name = os.path.basename(src_img_path)
    src_img_file_name_prefix = os.path.splitext(os.path.basename(src_img_path))[0]

    # Create prompt by task type
    task_type = config.model.de.task_type
    special_token = config.model.de.special_token
    class_name = config.data.class_name
    if task_type == "replace" or task_type == "add":
        src_prompt = "photo of a " + class_name
        dst_prompt = "photo of a " + special_token + " " + class_name
    else:
        src_prompt = None
        dst_prompt = None
    print("src_prompt: ", src_prompt)
    print("dst_prompt: ", dst_prompt)

    if task_type == "add":
        with open(bbox_file_path) as json_file:
            bbox_dict = json.load(json_file)
        img_edited, intermediate_images, generated_image_list = addition(
            src_img_path=src_img_path,
            obj_img_path=obj_img_path,
            obj_class=class_name,
            lang_sam_model=lang_sam_model,
            sd_model=sd_model,
            src_prompt=src_prompt,
            dst_prompt=dst_prompt,
            config=config,
            device=device,
            labeled_box=bbox_dict[src_img_file_name],
        )
    elif task_type == "replace":
        img_edited, intermediate_images, generated_image_list = replacement(
            src_img_path=src_img_path,
            obj_img_path=obj_img_path,
            obj_class=class_name,
            lang_sam_model=lang_sam_model,
            sd_model=sd_model,
            src_prompt=src_prompt,
            dst_prompt=dst_prompt,
            config=config,
            device=device,
        )
    else:
        raise NotImplementedError

    save_pil_image(
        pil_image=intermediate_images,
        dest_folder=os.path.join(
            config.experiment_result_path,
            config.data.db_folder_name,
            task_type
        ),
        filename=str(src_img_file_name_prefix) + "_edited",
        filetype="jpg",
    )

    # choose and save the image with the largest dino score
    obj_img_name = config.data.obj_img_file_name
    obj_img_path = os.path.join(
        config.data.db_dataset_path,
        config.data.db_folder_name,
        obj_img_name
    )
    src_img = Image.open(src_img_path).resize((512, 512))
    obj_img = Image.open(obj_img_path).resize((512, 512))
    dino_score_list_subject = evaluate_dino_score_list(obj_img, generated_image_list, device, dino_model)
    dino_score_list_background = evaluate_dino_score_list(src_img, generated_image_list, device, dino_model)
    dino_score_list_average = [round((dino_score_list_subject[x] + dino_score_list_background[x])/2, 3) for x in
                               range(len(dino_score_list_subject))]

    clipi_score_list_subject = evaluate_clipi_score_list(obj_img, generated_image_list, device, clip_model)
    clipi_score_list_background = evaluate_clipi_score_list(src_img, generated_image_list, device, clip_model)
    clipi_score_list_average = [round((clipi_score_list_subject[x] + clipi_score_list_background[x])/2, 3) for x in
                               range(len(clipi_score_list_subject))]


    max_dino_subject = max(dino_score_list_subject)
    max_dino_background = max(dino_score_list_background)
    max_dino_avg = max(dino_score_list_average)

    max_clipi_subject = max(clipi_score_list_subject)
    max_clipi_background = max(clipi_score_list_background)
    max_clipi_avg = max(clipi_score_list_average)

    max_index_dino_subject = dino_score_list_subject.index(max_dino_subject)
    max_index_dino_background = dino_score_list_background.index(max_dino_background)
    max_index_dino_avg = dino_score_list_average.index(max_dino_avg)

    max_index_clipi_subject = clipi_score_list_subject.index(max_clipi_subject)
    max_index_clipi_background = clipi_score_list_background.index(max_clipi_background)
    max_index_clipi_avg = clipi_score_list_average.index(max_clipi_avg)

    print("The best result of dino-subject is at {} th iteration with dino score {}".format(max_index_dino_subject + 1,
                                                                                            max_dino_subject))
    print("The best result of dino-background is at {} th iteration with dino score {}".format(max_index_dino_background + 1,
                                                                                            max_dino_background))
    print("The best result of dino-avg is at {} th iteration with dino score {}".format(
        max_index_dino_avg + 1,
        max_dino_avg))


    print("The best result of clipi-subject is at {} th iteration with clipi score {}".format(max_index_clipi_subject + 1,
                                                                                            max_clipi_subject))
    print("The best result of clipi-background is at {} th iteration with clipi score {}".format(max_index_clipi_background + 1,
                                                                                            max_clipi_background))
    print("The best result of clipi-avg is at {} th iteration with clipi score {}".format(
        max_index_clipi_avg + 1,
        max_clipi_avg))

    # ====== Max: Haven't saved clip-i score
    final_result = [src_img, generated_image_list[max_index_dino_avg]]
    final_result = get_concat_pil_images(final_result, direction="h")
    save_pil_image(
        pil_image=final_result,
        dest_folder=os.path.join(
            config.experiment_result_path,
            config.data.db_folder_name,
            task_type
        ),
        filename=str(src_img_file_name_prefix) + "_edited_final",
        filetype="jpg",
    )
    return max_dino_background, max_dino_subject, max_dino_avg, max_clipi_background, max_clipi_subject, max_clipi_avg, intermediate_images


def edit_all_images_from_class(config):
    # Load model
    sd_ckpt_path = config.model.sd.ckpt_path
    sd_conf = OmegaConf.load(config.model.sd.conf_path)
    sd_model = load_model_from_config(sd_conf, sd_ckpt_path)
    device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
    sd_model = sd_model.to(device)
    fidelity = VITs16(device)
    clip_model = CLIP(device)
    langSAM = LangSAM()

    # load all base images belonging to the class
    task_type = config.model.de.task_type
    class_name = config.data.class_name
    folder_name = config.data.db_folder_name
    obj_img_file_name = config.data.obj_img_file_name
    src_img_data_folder_path = config.data.src_img_data_folder_path
    all_files_path = os.path.join(src_img_data_folder_path, config.benchmark_folder_name, "*")
    print("Looking for images in path: ", all_files_path)
    all_files = sorted(glob.glob(all_files_path))

    if task_type == "edit":
        all_files = [img_name for img_name in all_files if img_name.endswith(obj_img_file_name)] * 3

    final_score_avg_list = []
    final_score_subject_list = []
    final_score_background_list = []

    final_score_avg_clipi_list = []
    final_score_subject_clipi_list = []
    final_score_background_clipi_list = []

    all_images_to_save = []

    print("Found files: ", all_files)

    for file_path in all_files:
        if file_path.endswith(".jpg"):
            print("Processing img:", file_path)
            bbox_file_path = os.path.join(
                config.data.src_img_data_folder_path,
                config.benchmark_folder_name,
                config.data.bbox_file_name,
            )
            obj_img_path = os.path.join(
                config.data.db_dataset_path,
                config.data.db_folder_name,
                config.data.obj_img_file_name,
            )
            max_dino_background, max_dino_subject, max_dino_avg, max_clipi_background, max_clipi_subject, max_clipi_avg, intermediate_images = dream_edit(
                src_img_path=file_path,
                device=device,
                config=config,
                obj_img_path=obj_img_path,
                bbox_file_path=bbox_file_path,
                sd_model=sd_model,
                lang_sam_model=langSAM,
                dino_model=fidelity,
                clip_model=clip_model
            )
            final_score_avg_list.append(max_dino_avg)
            final_score_subject_list.append(max_dino_subject)
            final_score_background_list.append(max_dino_background)
            final_score_avg_clipi_list.append(max_clipi_avg)
            final_score_subject_clipi_list.append(max_clipi_subject)
            final_score_background_clipi_list.append(max_clipi_background)
            all_images_to_save.append(intermediate_images)

    print("The average dino-subject score is: {}".format(sum(final_score_subject_list) / len(final_score_subject_list)))
    print("The average dino-background score is: {}".format(sum(final_score_background_list) / len(final_score_background_list)))
    print("The average dino-avg score is: {}".format(sum(final_score_avg_list) / len(final_score_avg_list)))
    print("The average clipi-subject score is: {}".format(sum(final_score_subject_clipi_list) / len(final_score_subject_clipi_list)))
    print("The average clipi-background score is: {}".format(
        sum(final_score_background_clipi_list) / len(final_score_background_clipi_list)))
    print("The average clipi-avg score is: {}".format(sum(final_score_avg_clipi_list) / len(final_score_avg_clipi_list)))
    all_images_concat = get_concat_pil_images(all_images_to_save, direction="v")
    save_pil_image(
        pil_image=all_images_concat,
        dest_folder=os.path.join(
            config.experiment_result_path,
            config.data.db_folder_name,
            task_type
        ),
        filename="result_all",
        filetype="jpg",
    )


def edit_single_img_from_class(config):
    # Load model
    sd_ckpt_path = config.model.sd.ckpt_path
    sd_conf = OmegaConf.load(config.model.sd.conf_path)
    sd_model = load_model_from_config(sd_conf, sd_ckpt_path)
    device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
    sd_model = sd_model.to(device)
    fidelity = VITs16(device)
    clip_model = CLIP(device)
    langSAM = LangSAM()
    src_img_path = os.path.join(
        config.data.src_img_data_folder_path,
        config.benchmark_folder_name,
        config.data.src_img_file_name,
    )
    bbox_file_path = os.path.join(
        config.data.src_img_data_folder_path,
        config.benchmark_folder_name,
        config.data.bbox_file_name,
    )
    obj_img_path = os.path.join(
        config.data.db_dataset_path,
        config.data.db_folder_name,
        config.data.obj_img_file_name,
    )
    max_dino_background, max_dino_subject, max_dino_avg, intermediate_images = dream_edit(
        src_img_path=src_img_path,
        device=device,
        config=config,
        obj_img_path=obj_img_path,
        bbox_file_path=bbox_file_path,
        sd_model=sd_model,
        lang_sam_model=langSAM,
        dino_model=fidelity,
        clip_model=clip_model
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        type=str,
        default="configs/stable-diffusion/v1-inference.yaml",
        help="path to config which constructs model",
    )
    parser.add_argument(
        "--edit_single_img_from_class",
        action="store_true",
        help="edit a single image from a class",
    )
    opt = parser.parse_args()
    config = OmegaConf.load(opt.config)
    if opt.edit_single_img_from_class:
        edit_single_img_from_class(config)
    else:
        edit_all_images_from_class(config)



import math

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint as cp
from annotator.uniformer.mmcv.cnn import build_conv_layer, build_norm_layer

from ..builder import BACKBONES
from ..utils import ResLayer
from .resnet import Bottleneck as _Bottleneck
from .resnet import ResNetV1d


class RSoftmax(nn.Module):
    """Radix Softmax module in ``SplitAttentionConv2d``.

    Args:
        radix (int): Radix of input.
        groups (int): Groups of input.
    """

    def __init__(self, radix, groups):
        super().__init__()
        self.radix = radix
        self.groups = groups

    def forward(self, x):
        batch = x.size(0)
        if self.radix > 1:
            x = x.view(batch, self.groups, self.radix, -1).transpose(1, 2)
            x = F.softmax(x, dim=1)
            x = x.reshape(batch, -1)
        else:
            x = torch.sigmoid(x)
        return x


class SplitAttentionConv2d(nn.Module):
    """Split-Attention Conv2d in ResNeSt.

    Args:
        in_channels (int): Same as nn.Conv2d.
        out_channels (int): Same as nn.Conv2d.
        kernel_size (int | tuple[int]): Same as nn.Conv2d.
        stride (int | tuple[int]): Same as nn.Conv2d.
        padding (int | tuple[int]): Same as nn.Conv2d.
        dilation (int | tuple[int]): Same as nn.Conv2d.
        groups (int): Same as nn.Conv2d.
        radix (int): Radix of SpltAtConv2d. Default: 2
        reduction_factor (int): Reduction factor of inter_channels. Default: 4.
        conv_cfg (dict): Config dict for convolution layer. Default: None,
            which means using conv2d.
        norm_cfg (dict): Config dict for normalization layer. Default: None.
        dcn (dict): Config dict for DCN. Default: None.
    """

    def __init__(self,
                 in_channels,
                 channels,
                 kernel_size,
                 stride=1,
                 padding=0,
                 dilation=1,
                 groups=1,
                 radix=2,
                 reduction_factor=4,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 dcn=None):
        super(SplitAttentionConv2d, self).__init__()
        inter_channels = max(in_channels * radix // reduction_factor, 32)
        self.radix = radix
        self.groups = groups
        self.channels = channels
        self.with_dcn = dcn is not None
        self.dcn = dcn
        fallback_on_stride = False
        if self.with_dcn:
            fallback_on_stride = self.dcn.pop('fallback_on_stride', False)
        if self.with_dcn and not fallback_on_stride:
            assert conv_cfg is None, 'conv_cfg must be None for DCN'
            conv_cfg = dcn
        self.conv = build_conv_layer(
            conv_cfg,
            in_channels,
            channels * radix,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups * radix,
            bias=False)
        self.norm0_name, norm0 = build_norm_layer(
            norm_cfg, channels * radix, postfix=0)
        self.add_module(self.norm0_name, norm0)
        self.relu = nn.ReLU(inplace=True)
        self.fc1 = build_conv_layer(
            None, channels, inter_channels, 1, groups=self.groups)
        self.norm1_name, norm1 = build_norm_layer(
            norm_cfg, inter_channels, postfix=1)
        self.add_module(self.norm1_name, norm1)
        self.fc2 = build_conv_layer(
            None, inter_channels, channels * radix, 1, groups=self.groups)
        self.rsoftmax = RSoftmax(radix, groups)

    @property
    def norm0(self):
        """nn.Module: the normalization layer named "norm0" """
        return getattr(self, self.norm0_name)

    @property
    def norm1(self):
        """nn.Module: the normalization layer named "norm1" """
        return getattr(self, self.norm1_name)

    def forward(self, x):
        x = self.conv(x)
        x = self.norm0(x)
        x = self.relu(x)

        batch, rchannel = x.shape[:2]
        batch = x.size(0)
        if self.radix > 1:
            splits = x.view(batch, self.radix, -1, *x.shape[2:])
            gap = splits.sum(dim=1)
        else:
            gap = x
        gap = F.adaptive_avg_pool2d(gap, 1)
        gap = self.fc1(gap)

        gap = self.norm1(gap)
        gap = self.relu(gap)

        atten = self.fc2(gap)
        atten = self.rsoftmax(atten).view(batch, -1, 1, 1)

        if self.radix > 1:
            attens = atten.view(batch, self.radix, -1, *atten.shape[2:])
            out = torch.sum(attens * splits, dim=1)
        else:
            out = atten * x
        return out.contiguous()


class Bottleneck(_Bottleneck):
    """Bottleneck block for ResNeSt.

    Args:
        inplane (int): Input planes of this block.
        planes (int): Middle planes of this block.
        groups (int): Groups of conv2.
        width_per_group (int): Width per group of conv2. 64x4d indicates
            ``groups=64, width_per_group=4`` and 32x8d indicates
            ``groups=32, width_per_group=8``.
        radix (int): Radix of SpltAtConv2d. Default: 2
        reduction_factor (int): Reduction factor of inter_channels in
            SplitAttentionConv2d. Default: 4.
        avg_down_stride (bool): Whether to use average pool for stride in
            Bottleneck. Default: True.
        kwargs (dict): Key word arguments for base class.
    """
    expansion = 4

    def __init__(self,
                 inplanes,
                 planes,
                 groups=1,
                 base_width=4,
                 base_channels=64,
                 radix=2,
                 reduction_factor=4,
                 avg_down_stride=True,
                 **kwargs):
        """Bottleneck block for ResNeSt."""
        super(Bottleneck, self).__init__(inplanes, planes, **kwargs)

        if groups == 1:
            width = self.planes
        else:
            width = math.floor(self.planes *
                               (base_width / base_channels)) * groups

        self.avg_down_stride = avg_down_stride and self.conv2_stride > 1

        self.norm1_name, norm1 = build_norm_layer(
            self.norm_cfg, width, postfix=1)
        self.norm3_name, norm3 = build_norm_layer(
            self.norm_cfg, self.planes * self.expansion, postfix=3)

        self.conv1 = build_conv_layer(
            self.conv_cfg,
            self.inplanes,
            width,
            kernel_size=1,
            stride=self.conv1_stride,
            bias=False)
        self.add_module(self.norm1_name, norm1)
        self.with_modulated_dcn = False
        self.conv2 = SplitAttentionConv2d(
            width,
            width,
            kernel_size=3,
            stride=1 if self.avg_down_stride else self.conv2_stride,
            padding=self.dilation,
            dilation=self.dilation,
            groups=groups,
            radix=radix,
            reduction_factor=reduction_factor,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            dcn=self.dcn)
        delattr(self, self.norm2_name)

        if self.avg_down_stride:
            self.avd_layer = nn.AvgPool2d(3, self.conv2_stride, padding=1)

        self.conv3 = build_conv_layer(
            self.conv_cfg,
            width,
            self.planes * self.expansion,
            kernel_size=1,
            bias=False)
        self.add_module(self.norm3_name, norm3)

    def forward(self, x):

        def _inner_forward(x):
            identity = x

            out = self.conv1(x)
            out = self.norm1(out)
            out = self.relu(out)

            if self.with_plugins:
                out = self.forward_plugin(out, self.after_conv1_plugin_names)

            out = self.conv2(out)

            if self.avg_down_stride:
                out = self.avd_layer(out)

            if self.with_plugins:
                out = self.forward_plugin(out, self.after_conv2_plugin_names)

            out = self.conv3(out)
            out = self.norm3(out)

            if self.with_plugins:
                out = self.forward_plugin(out, self.after_conv3_plugin_names)

            if self.downsample is not None:
                identity = self.downsample(x)

            out += identity

            return out

        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(_inner_forward, x)
        else:
            out = _inner_forward(x)

        out = self.relu(out)

        return out


@BACKBONES.register_module()
class ResNeSt(ResNetV1d):
    """ResNeSt backbone.

    Args:
        groups (int): Number of groups of Bottleneck. Default: 1
        base_width (int): Base width of Bottleneck. Default: 4
        radix (int): Radix of SpltAtConv2d. Default: 2
        reduction_factor (int): Reduction factor of inter_channels in
            SplitAttentionConv2d. Default: 4.
        avg_down_stride (bool): Whether to use average pool for stride in
            Bottleneck. Default: True.
        kwargs (dict): Keyword arguments for ResNet.
    """

    arch_settings = {
        50: (Bottleneck, (3, 4, 6, 3)),
        101: (Bottleneck, (3, 4, 23, 3)),
        152: (Bottleneck, (3, 8, 36, 3)),
        200: (Bottleneck, (3, 24, 36, 3))
    }

    def __init__(self,
                 groups=1,
                 base_width=4,
                 radix=2,
                 reduction_factor=4,
                 avg_down_stride=True,
                 **kwargs):
        self.groups = groups
        self.base_width = base_width
        self.radix = radix
        self.reduction_factor = reduction_factor
        self.avg_down_stride = avg_down_stride
        super(ResNeSt, self).__init__(**kwargs)

    def make_res_layer(self, **kwargs):
        """Pack all blocks in a stage into a ``ResLayer``."""
        return ResLayer(
            groups=self.groups,
            base_width=self.base_width,
            base_channels=self.base_channels,
            radix=self.radix,
            reduction_factor=self.reduction_factor,
            avg_down_stride=self.avg_down_stride,
            **kwargs)



"""Clip index is a tool to index clip embeddings using autofaiss"""

import fire
import os
from shutil import copytree
import logging


LOGGER = logging.getLogger(__name__)


def quantize(emb_folder, index_folder, index_name, max_index_memory_usage, current_memory_available, nb_cores):
    """calls autofaiss to build an index"""

    from autofaiss import build_index  # pylint: disable=import-outside-toplevel

    try:
        LOGGER.debug(f"starting index {index_name}")
        if os.path.exists(emb_folder):
            LOGGER.debug(
                f"embedding path exist, building index {index_name}"
                f"using embeddings {emb_folder} ; saving in {index_folder}"
            )
            build_index(
                embeddings=emb_folder,
                index_path=index_folder + "/" + index_name + ".index",
                index_infos_path=index_folder + "/" + index_name + ".json",
                max_index_memory_usage=max_index_memory_usage,
                current_memory_available=current_memory_available,
                nb_cores=nb_cores,
            )
            LOGGER.debug(f"index {index_name} done")
    except Exception as e:  # pylint: disable=broad-except
        LOGGER.exception(f"index {index_name} failed")
        raise e


def clip_index(
    embeddings_folder,
    index_folder,
    max_index_memory_usage="4G",
    current_memory_available="16G",
    copy_metadata=True,
    image_subfolder="img_emb",
    text_subfolder="text_emb",
    nb_cores=None,
):
    """indexes clip embeddings using autofaiss"""
    quantize(
        embeddings_folder + "/" + image_subfolder,
        index_folder,
        "image",
        max_index_memory_usage,
        current_memory_available,
        nb_cores,
    )
    quantize(
        embeddings_folder + "/" + text_subfolder,
        index_folder,
        "text",
        max_index_memory_usage,
        current_memory_available,
        nb_cores,
    )
    if copy_metadata:
        copytree(embeddings_folder + "/metadata", index_folder + "/metadata")


if __name__ == "__main__":
    fire.Fire(clip_index)
